@@ -4,6 +4,8 @@
 
 part of dart._http;
 
+final _digitsValidator = RegExp(r"^\d+$");
+
 class _HttpHeaders implements HttpHeaders {
   final Map<String, List<String>> _headers;
   // The original header names keyed by the lowercase header names.
@@ -21,11 +23,12 @@ class _HttpHeaders implements HttpHeaders {
 
   final int _defaultPortForScheme;
 
-  _HttpHeaders(this.protocolVersion,
-      {int defaultPortForScheme = HttpClient.defaultHttpPort,
-      _HttpHeaders? initialHeaders})
-      : _headers = new HashMap<String, List<String>>(),
-        _defaultPortForScheme = defaultPortForScheme {
+  _HttpHeaders(
+    this.protocolVersion, {
+    int defaultPortForScheme = HttpClient.defaultHttpPort,
+    _HttpHeaders? initialHeaders,
+  }) : _headers = HashMap<String, List<String>>(),
+       _defaultPortForScheme = defaultPortForScheme {
     if (initialHeaders != null) {
       initialHeaders._headers.forEach((name, value) => _headers[name] = value);
       _contentLength = initialHeaders._contentLength;
@@ -48,7 +51,7 @@ class _HttpHeaders implements HttpHeaders {
     if (values == null) return null;
     assert(values.isNotEmpty);
     if (values.length > 1) {
-      throw new HttpException("More than one value for header $name");
+      throw HttpException("More than one value for header $name");
     }
     return values[0];
   }
@@ -99,7 +102,7 @@ class _HttpHeaders implements HttpHeaders {
     List<String>? values = _headers[name];
     if (values != null) {
       values.remove(_valueToString(value));
-      if (values.length == 0) {
+      if (values.isEmpty) {
         _headers.remove(name);
         _originalHeaderNames?.remove(name);
       }
@@ -116,7 +119,7 @@ class _HttpHeaders implements HttpHeaders {
     _originalHeaderNames?.remove(name);
   }
 
-  void forEach(void action(String name, List<String> values)) {
+  void forEach(void Function(String name, List<String> values) action) {
     _headers.forEach((String name, List<String> values) {
       String originalName = _originalHeaderName(name);
       action(originalName, values);
@@ -139,9 +142,10 @@ class _HttpHeaders implements HttpHeaders {
         remove(HttpHeaders.connectionHeader, "close");
       } else {
         if (_contentLength < 0) {
-          throw new HttpException(
-              "Trying to set 'Connection: Keep-Alive' on HTTP 1.0 headers with "
-              "no ContentLength");
+          throw HttpException(
+            "Trying to set 'Connection: Keep-Alive' on HTTP 1.0 headers with "
+            "no ContentLength",
+          );
         }
         add(originalName, "keep-alive", preserveHeaderCase: true);
       }
@@ -162,9 +166,10 @@ class _HttpHeaders implements HttpHeaders {
     if (protocolVersion == "1.0" &&
         persistentConnection &&
         contentLength == -1) {
-      throw new HttpException(
-          "Trying to clear ContentLength on HTTP 1.0 headers with "
-          "'Connection: Keep-Alive' set");
+      throw HttpException(
+        "Trying to clear ContentLength on HTTP 1.0 headers with "
+        "'Connection: Keep-Alive' set",
+      );
     }
     if (_contentLength == contentLength) return;
     _contentLength = contentLength;
@@ -184,8 +189,9 @@ class _HttpHeaders implements HttpHeaders {
   void set chunkedTransferEncoding(bool chunkedTransferEncoding) {
     _checkMutable();
     if (chunkedTransferEncoding && protocolVersion == "1.0") {
-      throw new HttpException(
-          "Trying to set 'Transfer-Encoding: Chunked' on HTTP 1.0 headers");
+      throw HttpException(
+        "Trying to set 'Transfer-Encoding: Chunked' on HTTP 1.0 headers",
+      );
     }
     if (chunkedTransferEncoding == _chunkedTransferEncoding) return;
     if (chunkedTransferEncoding) {
@@ -373,12 +379,18 @@ class _HttpHeaders implements HttpHeaders {
 
   void _addContentLength(String name, value) {
     if (value is int) {
-      contentLength = value;
+      if (value < 0) {
+        throw HttpException("Content-Length must contain only digits");
+      }
     } else if (value is String) {
-      contentLength = int.parse(value);
+      if (!_digitsValidator.hasMatch(value)) {
+        throw HttpException("Content-Length must contain only digits");
+      }
+      value = int.parse(value);
     } else {
-      throw new HttpException("Unexpected type for header named $name");
+      throw HttpException("Unexpected type for header named $name");
     }
+    contentLength = value;
   }
 
   void _addTransferEncoding(String name, value) {
@@ -395,7 +407,7 @@ class _HttpHeaders implements HttpHeaders {
     } else if (value is String) {
       _set(HttpHeaders.dateHeader, value);
     } else {
-      throw new HttpException("Unexpected type for header named $name");
+      throw HttpException("Unexpected type for header named $name");
     }
   }
 
@@ -405,7 +417,7 @@ class _HttpHeaders implements HttpHeaders {
     } else if (value is String) {
       _set(HttpHeaders.expiresHeader, value);
     } else {
-      throw new HttpException("Unexpected type for header named $name");
+      throw HttpException("Unexpected type for header named $name");
     }
   }
 
@@ -415,14 +427,19 @@ class _HttpHeaders implements HttpHeaders {
     } else if (value is String) {
       _set(HttpHeaders.ifModifiedSinceHeader, value);
     } else {
-      throw new HttpException("Unexpected type for header named $name");
+      throw HttpException("Unexpected type for header named $name");
     }
   }
 
   void _addHost(String name, value) {
     if (value is String) {
-      int pos = value.indexOf(":");
-      if (pos == -1) {
+      // value.indexOf will only work for ipv4, ipv6 which has multiple : in its
+      // host part needs lastIndexOf
+      int pos = value.lastIndexOf(":");
+      // According to RFC 3986, section 3.2.2, host part of ipv6 address must be
+      // enclosed by square brackets.
+      // https://serverfault.com/questions/205793/how-can-one-distinguish-the-host-and-the-port-in-an-ipv6-url
+      if (pos == -1 || value.startsWith("[") && value.endsWith("]")) {
         _host = value;
         _port = HttpClient.defaultHttpPort;
       } else {
@@ -443,11 +460,11 @@ class _HttpHeaders implements HttpHeaders {
       }
       _set(HttpHeaders.hostHeader, value);
     } else {
-      throw new HttpException("Unexpected type for header named $name");
+      throw HttpException("Unexpected type for header named $name");
     }
   }
 
-  void _addConnection(String name, value) {
+  void _addConnection(String name, String value) {
     var lowerCaseValue = value.toLowerCase();
     if (lowerCaseValue == 'close') {
       _persistentConnection = false;
@@ -482,7 +499,7 @@ class _HttpHeaders implements HttpHeaders {
   }
 
   void _checkMutable() {
-    if (!_mutable) throw new HttpException("HTTP headers are not mutable");
+    if (!_mutable) throw HttpException("HTTP headers are not mutable");
   }
 
   void _updateHostHeader() {
@@ -503,8 +520,20 @@ class _HttpHeaders implements HttpHeaders {
     _mutable = false;
   }
 
-  void _build(BytesBuilder builder) {
+  void _build(BytesBuilder builder, {bool skipZeroContentLength = false}) {
+    // per https://tools.ietf.org/html/rfc7230#section-3.3.2
+    // A user agent SHOULD NOT send a
+    // Content-Length header field when the request message does not
+    // contain a payload body and the method semantics do not anticipate
+    // such a body.
+    String? ignoreHeader =
+        _contentLength == 0 && skipZeroContentLength
+            ? HttpHeaders.contentLengthHeader
+            : null;
     _headers.forEach((String name, List<String> values) {
+      if (ignoreHeader == name) {
+        return;
+      }
       String originalName = _originalHeaderName(name);
       bool fold = _foldHeader(name);
       var nameData = originalName.codeUnits;
@@ -532,17 +561,22 @@ class _HttpHeaders implements HttpHeaders {
   }
 
   String toString() {
-    StringBuffer sb = new StringBuffer();
+    StringBuffer sb = StringBuffer();
     _headers.forEach((String name, List<String> values) {
       String originalName = _originalHeaderName(name);
-      sb..write(originalName)..write(": ");
+      sb
+        ..write(originalName)
+        ..write(": ");
       bool fold = _foldHeader(name);
       for (int i = 0; i < values.length; i++) {
         if (i > 0) {
           if (fold) {
             sb.write(", ");
           } else {
-            sb..write("\n")..write(originalName)..write(": ");
+            sb
+              ..write("\n")
+              ..write(originalName)
+              ..write(": ");
           }
         }
         sb.write(values[i]);
@@ -604,7 +638,7 @@ class _HttpHeaders implements HttpHeaders {
         skipWS();
         String value = parseValue();
         try {
-          cookies.add(new _Cookie(name, value));
+          cookies.add(_Cookie(name, value));
         } catch (_) {
           // Skip it, invalid cookie data.
         }
@@ -619,7 +653,9 @@ class _HttpHeaders implements HttpHeaders {
 
     List<String>? values = _headers[HttpHeaders.cookieHeader];
     if (values != null) {
-      values.forEach((headerValue) => parseCookieString(headerValue));
+      for (var headerValue in values) {
+        parseCookieString(headerValue);
+      }
     }
     return cookies;
   }
@@ -627,8 +663,11 @@ class _HttpHeaders implements HttpHeaders {
   static String _validateField(String field) {
     for (var i = 0; i < field.length; i++) {
       if (!_HttpParser._isTokenChar(field.codeUnitAt(i))) {
-        throw new FormatException(
-            "Invalid HTTP header field name: ${json.encode(field)}", field, i);
+        throw FormatException(
+          "Invalid HTTP header field name: ${json.encode(field)}",
+          field,
+          i,
+        );
       }
     }
     return field.toLowerCase();
@@ -636,18 +675,20 @@ class _HttpHeaders implements HttpHeaders {
 
   static Object _validateValue(Object value) {
     if (value is! String) return value;
-    for (var i = 0; i < (value as String).length; i++) {
-      if (!_HttpParser._isValueChar((value as String).codeUnitAt(i))) {
-        throw new FormatException(
-            "Invalid HTTP header field value: ${json.encode(value)}", value, i);
+    for (var i = 0; i < (value).length; i++) {
+      if (!_HttpParser._isValueChar((value).codeUnitAt(i))) {
+        throw FormatException(
+          "Invalid HTTP header field value: ${json.encode(value)}",
+          value,
+          i,
+        );
       }
     }
     return value;
   }
 
   String _originalHeaderName(String name) {
-    // TODO: Update syntax to_originalHeaderNames?[name].
-    return _originalHeaderNames?.[name] ?? name;
+    return _originalHeaderNames?[name] ?? name;
   }
 }
 
@@ -660,16 +701,18 @@ class _HeaderValue implements HeaderValue {
     // TODO(40614): Remove once non-nullability is sound.
     Map<String, String?>? nullableParameters = parameters;
     if (nullableParameters != null && nullableParameters.isNotEmpty) {
-      _parameters = new HashMap<String, String?>.from(nullableParameters);
+      _parameters = HashMap<String, String?>.from(nullableParameters);
     }
   }
 
-  static _HeaderValue parse(String value,
-      {String parameterSeparator = ";",
-      String? valueSeparator,
-      bool preserveBackslash = false}) {
+  static _HeaderValue parse(
+    String value, {
+    String parameterSeparator = ";",
+    String? valueSeparator,
+    bool preserveBackslash = false,
+  }) {
     // Parse the string.
-    var result = new _HeaderValue();
+    var result = _HeaderValue();
     result._parse(value, parameterSeparator, valueSeparator, preserveBackslash);
     return result;
   }
@@ -686,12 +729,10 @@ class _HeaderValue implements HeaderValue {
     if (token.isEmpty) {
       return false;
     }
-    final delimiters = "\"(),/:;<=>?@[\]{}";
+    final delimiters = "\"(),/:;<=>?@[]{}";
     for (int i = 0; i < token.length; i++) {
       int codeUnit = token.codeUnitAt(i);
-      if (codeUnit <= 32 ||
-          codeUnit >= 127 ||
-          delimiters.indexOf(token[i]) >= 0) {
+      if (codeUnit <= 32 || codeUnit >= 127 || delimiters.contains(token[i])) {
         return false;
       }
     }
@@ -699,12 +740,14 @@ class _HeaderValue implements HeaderValue {
   }
 
   String toString() {
-    StringBuffer sb = new StringBuffer();
+    StringBuffer sb = StringBuffer();
     sb.write(_value);
-    var parameters = this._parameters;
-    if (parameters != null && parameters.length > 0) {
+    var parameters = _parameters;
+    if (parameters != null && parameters.isNotEmpty) {
       parameters.forEach((String name, String? value) {
-        sb..write("; ")..write(name);
+        sb
+          ..write("; ")
+          ..write(name);
         if (value != null) {
           sb.write("=");
           if (_isToken(value)) {
@@ -716,13 +759,15 @@ class _HeaderValue implements HeaderValue {
               // Can use codeUnitAt here instead.
               int codeUnit = value.codeUnitAt(i);
               if (codeUnit == 92 /* backslash */ ||
-                  codeUnit == 34 /* double quote */) {
+                  codeUnit == 34 /* double quote */ ) {
                 sb.write(value.substring(start, i));
                 sb.write(r'\');
                 start = i;
               }
             }
-            sb..write(value.substring(start))..write('"');
+            sb
+              ..write(value.substring(start))
+              ..write('"');
           }
         }
       });
@@ -730,8 +775,12 @@ class _HeaderValue implements HeaderValue {
     return sb.toString();
   }
 
-  void _parse(String s, String parameterSeparator, String? valueSeparator,
-      bool preserveBackslash) {
+  void _parse(
+    String s,
+    String parameterSeparator,
+    String? valueSeparator,
+    bool preserveBackslash,
+  ) {
     int index = 0;
 
     bool done() => index == s.length;
@@ -750,7 +799,8 @@ class _HeaderValue implements HeaderValue {
         if (char == " " ||
             char == "\t" ||
             char == valueSeparator ||
-            char == parameterSeparator) break;
+            char == parameterSeparator)
+          break;
         index++;
       }
       return s.substring(start, index);
@@ -758,7 +808,7 @@ class _HeaderValue implements HeaderValue {
 
     void expect(String expected) {
       if (done() || s[index] != expected) {
-        throw new HttpException("Failed to parse header value");
+        throw HttpException("Failed to parse header value");
       }
       index++;
     }
@@ -782,7 +832,8 @@ class _HeaderValue implements HeaderValue {
               char == "\t" ||
               char == "=" ||
               char == parameterSeparator ||
-              char == valueSeparator) break;
+              char == valueSeparator)
+            break;
           index++;
         }
         return s.substring(start, index).toLowerCase();
@@ -791,13 +842,13 @@ class _HeaderValue implements HeaderValue {
       String parseParameterValue() {
         if (!done() && s[index] == "\"") {
           // Parse quoted value.
-          StringBuffer sb = new StringBuffer();
+          StringBuffer sb = StringBuffer();
           index++;
           while (!done()) {
             var char = s[index];
             if (char == "\\") {
               if (index + 1 == s.length) {
-                throw new HttpException("Failed to parse header value");
+                throw HttpException("Failed to parse header value");
               }
               if (preserveBackslash && s[index + 1] != "\"") {
                 sb.write(char);
@@ -811,7 +862,7 @@ class _HeaderValue implements HeaderValue {
             sb.write(char);
             index++;
           }
-          throw new HttpException("Failed to parse header value");
+          throw HttpException("Failed to parse header value");
         } else {
           // Parse non-quoted value.
           return parseValue();
@@ -856,11 +907,14 @@ class _ContentType extends _HeaderValue implements ContentType {
   String _primaryType = "";
   String _subType = "";
 
-  _ContentType(String primaryType, String subType, String? charset,
-      Map<String, String?> parameters)
-      : _primaryType = primaryType,
-        _subType = subType,
-        super("") {
+  _ContentType(
+    String primaryType,
+    String subType,
+    String? charset,
+    Map<String, String?> parameters,
+  ) : _primaryType = primaryType,
+      _subType = subType,
+      super("") {
     // TODO(40614): Remove once non-nullability is sound.
     String emptyIfNull(String? string) => string ?? "";
     _primaryType = emptyIfNull(_primaryType);
@@ -886,7 +940,7 @@ class _ContentType extends _HeaderValue implements ContentType {
   _ContentType._();
 
   static _ContentType parse(String value) {
-    var result = new _ContentType._();
+    var result = _ContentType._();
     result._parse(value, ";", null, false);
     int index = result._value.indexOf("/");
     if (index == -1 || index == (result._value.length - 1)) {
@@ -914,18 +968,26 @@ class _Cookie implements Cookie {
   DateTime? expires;
   int? maxAge;
   String? domain;
-  String? path;
+  String? _path;
   bool httpOnly = false;
   bool secure = false;
+  SameSite? sameSite;
 
   _Cookie(String name, String value)
-      : _name = _validateName(name),
-        _value = _validateValue(value),
-        httpOnly = true;
+    : _name = _validateName(name),
+      _value = _validateValue(value),
+      httpOnly = true;
 
   String get name => _name;
   String get value => _value;
 
+  String? get path => _path;
+
+  set path(String? newPath) {
+    _validatePath(newPath);
+    _path = newPath;
+  }
+
   set name(String newName) {
     _validateName(newName);
     _name = newName;
@@ -936,9 +998,7 @@ class _Cookie implements Cookie {
     _value = newValue;
   }
 
-  _Cookie.fromSetCookieValue(String value)
-      : _name = "",
-        _value = "" {
+  _Cookie.fromSetCookieValue(String value) : _name = "", _value = "" {
     // Parse the 'set-cookie' header value.
     _parseSetCookieValue(value);
   }
@@ -1005,14 +1065,24 @@ class _Cookie implements Cookie {
           httpOnly = true;
         } else if (name == "secure") {
           secure = true;
+        } else if (name == "samesite") {
+          sameSite = switch (value) {
+            "lax" => SameSite.lax,
+            "none" => SameSite.none,
+            "strict" => SameSite.strict,
+            _ =>
+              throw HttpException(
+                'SameSite value should be one of Lax, Strict or None.',
+              ),
+          };
         }
         if (!done()) index++; // Skip the ; character
       }
     }
 
     _name = _validateName(parseName());
-    if (done() || _name.length == 0) {
-      throw new HttpException("Failed to parse header value [$s]");
+    if (done() || _name.isEmpty) {
+      throw HttpException("Failed to parse header value [$s]");
     }
     index++; // Skip the = character.
     _value = _validateValue(parseValue());
@@ -1022,28 +1092,41 @@ class _Cookie implements Cookie {
   }
 
   String toString() {
-    StringBuffer sb = new StringBuffer();
-    sb..write(_name)..write("=")..write(_value);
+    StringBuffer sb = StringBuffer();
+    sb
+      ..write(_name)
+      ..write("=")
+      ..write(_value);
     var expires = this.expires;
     if (expires != null) {
-      sb..write("; Expires=")..write(HttpDate.format(expires));
+      sb
+        ..write("; Expires=")
+        ..write(HttpDate.format(expires));
     }
     if (maxAge != null) {
-      sb..write("; Max-Age=")..write(maxAge);
+      sb
+        ..write("; Max-Age=")
+        ..write(maxAge);
     }
     if (domain != null) {
-      sb..write("; Domain=")..write(domain);
+      sb
+        ..write("; Domain=")
+        ..write(domain);
     }
     if (path != null) {
-      sb..write("; Path=")..write(path);
+      sb
+        ..write("; Path=")
+        ..write(path);
     }
     if (secure) sb.write("; Secure");
     if (httpOnly) sb.write("; HttpOnly");
+    if (sameSite != null) sb.write("; $sameSite");
+
     return sb.toString();
   }
 
   static String _validateName(String newName) {
-    const separators = const [
+    const separators = [
       "(",
       ")",
       "<",
@@ -1060,25 +1143,26 @@ class _Cookie implements Cookie {
       "?",
       "=",
       "{",
-      "}"
+      "}",
     ];
-    if (newName == null) throw new ArgumentError.notNull("name");
+    if (newName == null) throw ArgumentError.notNull("name");
     for (int i = 0; i < newName.length; i++) {
       int codeUnit = newName.codeUnitAt(i);
       if (codeUnit <= 32 ||
           codeUnit >= 127 ||
-          separators.indexOf(newName[i]) >= 0) {
-        throw new FormatException(
-            "Invalid character in cookie name, code unit: '$codeUnit'",
-            newName,
-            i);
+          separators.contains(newName[i])) {
+        throw FormatException(
+          "Invalid character in cookie name, code unit: '$codeUnit'",
+          newName,
+          i,
+        );
       }
     }
     return newName;
   }
 
   static String _validateValue(String newValue) {
-    if (newValue == null) throw new ArgumentError.notNull("value");
+    if (newValue == null) throw ArgumentError.notNull("value");
     // Per RFC 6265, consider surrounding "" as part of the value, but otherwise
     // double quotes are not allowed.
     int start = 0;
@@ -1097,12 +1181,29 @@ class _Cookie implements Cookie {
           (codeUnit >= 0x2D && codeUnit <= 0x3A) ||
           (codeUnit >= 0x3C && codeUnit <= 0x5B) ||
           (codeUnit >= 0x5D && codeUnit <= 0x7E))) {
-        throw new FormatException(
-            "Invalid character in cookie value, code unit: '$codeUnit'",
-            newValue,
-            i);
+        throw FormatException(
+          "Invalid character in cookie value, code unit: '$codeUnit'",
+          newValue,
+          i,
+        );
       }
     }
     return newValue;
   }
+
+  static void _validatePath(String? path) {
+    if (path == null) return;
+    for (int i = 0; i < path.length; i++) {
+      int codeUnit = path.codeUnitAt(i);
+      // According to RFC 6265, semicolon and controls should not occur in the
+      // path.
+      // path-value = <any CHAR except CTLs or ";">
+      // CTLs = %x00-1F / %x7F
+      if (codeUnit < 0x20 || codeUnit >= 0x7f || codeUnit == 0x3b /*;*/ ) {
+        throw FormatException(
+          "Invalid character in cookie path, code unit: '$codeUnit'",
+        );
+      }
+    }
+  }
 }
