@@ -4,19 +4,27 @@
 
 // Patch file for the dart:async library.
 
-import 'dart:_internal' show patch;
+import 'dart:_internal' show patch, unsafeCast;
 import 'dart:_js_helper'
     show
         ExceptionAndStackTrace,
         convertDartClosureToJS,
         getTraceFromException,
+        Primitives,
         requiresPreamble,
-        wrapException,
         unwrapException;
 
-import 'dart:_foreign_helper' show JS, JS_GET_FLAG;
+import 'dart:_foreign_helper'
+    show JS, JS_FALSE, JS_RAW_EXCEPTION, RAW_DART_FUNCTION_REF;
 
-import 'dart:_async_await_error_codes' as async_error_codes;
+import 'dart:_async_status_codes' as async_status_codes;
+
+@patch
+void _trySetStackTrace(Object error, StackTrace stackTrace) {
+  if (error is Error) {
+    Primitives.trySetStackTrace(error, stackTrace);
+  }
+}
 
 @patch
 class _AsyncRun {
@@ -47,8 +55,11 @@ class _AsyncRun {
         f!();
       }
 
-      var observer = JS('', 'new self.MutationObserver(#)',
-          convertDartClosureToJS(internalCallback, 1));
+      var observer = JS(
+        '',
+        'new self.MutationObserver(#)',
+        convertDartClosureToJS(internalCallback, 1),
+      );
       JS('', '#.observe(#, { childList: true })', observer, div);
 
       return (void callback()) {
@@ -56,8 +67,15 @@ class _AsyncRun {
         storedCallback = callback;
         // Because of a broken shadow-dom polyfill we have to change the
         // children instead a cheap property.
-        JS('', '#.firstChild ? #.removeChild(#): #.appendChild(#)', div, div,
-            span, div, span);
+        JS(
+          '',
+          '#.firstChild ? #.removeChild(#): #.appendChild(#)',
+          div,
+          div,
+          span,
+          div,
+          span,
+        );
       };
     } else if (JS('', 'self.setImmediate') != null) {
       return _scheduleImmediateWithSetImmediate;
@@ -71,8 +89,11 @@ class _AsyncRun {
       callback();
     }
 
-    JS('void', 'self.scheduleImmediate(#)',
-        convertDartClosureToJS(internalCallback, 0));
+    JS(
+      'void',
+      'self.scheduleImmediate(#)',
+      convertDartClosureToJS(internalCallback, 0),
+    );
   }
 
   static void _scheduleImmediateWithSetImmediate(void callback()) {
@@ -80,8 +101,11 @@ class _AsyncRun {
       callback();
     }
 
-    JS('void', 'self.setImmediate(#)',
-        convertDartClosureToJS(internalCallback, 0));
+    JS(
+      'void',
+      'self.setImmediate(#)',
+      convertDartClosureToJS(internalCallback, 0),
+    );
   }
 
   static void _scheduleImmediateWithTimer(void callback()) {
@@ -95,15 +119,17 @@ class Timer {
   static Timer _createTimer(Duration duration, void callback()) {
     int milliseconds = duration.inMilliseconds;
     if (milliseconds < 0) milliseconds = 0;
-    return new _TimerImpl(milliseconds, callback);
+    return _TimerImpl(milliseconds, callback);
   }
 
   @patch
   static Timer _createPeriodicTimer(
-      Duration duration, void callback(Timer timer)) {
+    Duration duration,
+    void callback(Timer timer),
+  ) {
     int milliseconds = duration.inMilliseconds;
     if (milliseconds < 0) milliseconds = 0;
-    return new _TimerImpl.periodic(milliseconds, callback);
+    return _TimerImpl.periodic(milliseconds, callback);
   }
 }
 
@@ -120,35 +146,40 @@ class _TimerImpl implements Timer {
         callback();
       }
 
-      _handle = JS('int', 'self.setTimeout(#, #)',
-          convertDartClosureToJS(internalCallback, 0), milliseconds);
+      _handle = JS(
+        'int',
+        'self.setTimeout(#, #)',
+        convertDartClosureToJS(internalCallback, 0),
+        milliseconds,
+      );
     } else {
-      throw new UnsupportedError('`setTimeout()` not found.');
+      throw UnsupportedError('`setTimeout()` not found.');
     }
   }
 
   _TimerImpl.periodic(int milliseconds, void callback(Timer timer))
-      : _once = false {
+    : _once = false {
     if (_hasTimer()) {
       int start = JS('int', 'Date.now()');
       _handle = JS(
-          'int',
-          'self.setInterval(#, #)',
-          convertDartClosureToJS(() {
-            int tick = this._tick + 1;
-            if (milliseconds > 0) {
-              int end = JS('int', 'Date.now()');
-              int duration = end - start;
-              if (duration > (tick + 1) * milliseconds) {
-                tick = duration ~/ milliseconds;
-              }
+        'int',
+        'self.setInterval(#, #)',
+        convertDartClosureToJS(() {
+          int tick = this._tick + 1;
+          if (milliseconds > 0) {
+            int end = JS('int', 'Date.now()');
+            int duration = end - start;
+            if (duration > (tick + 1) * milliseconds) {
+              tick = duration ~/ milliseconds;
             }
-            this._tick = tick;
-            callback(this);
-          }, 0),
-          milliseconds);
+          }
+          this._tick = tick;
+          callback(this);
+        }, 0),
+        milliseconds,
+      );
     } else {
-      throw new UnsupportedError('Periodic timer.');
+      throw UnsupportedError('Periodic timer.');
     }
   }
 
@@ -169,7 +200,7 @@ class _TimerImpl implements Timer {
       }
       _handle = null;
     } else {
-      throw new UnsupportedError('Canceling a timer.');
+      throw UnsupportedError('Canceling a timer.');
     }
   }
 }
@@ -180,7 +211,7 @@ bool _hasTimer() {
 }
 
 class _AsyncAwaitCompleter<T> implements Completer<T> {
-  final _future = new _Future<T>();
+  final _future = _Future<T>();
   bool isSync;
 
   _AsyncAwaitCompleter() : isSync = false;
@@ -194,11 +225,7 @@ class _AsyncAwaitCompleter<T> implements Completer<T> {
       assert(!_future._isComplete);
       _future._chainFuture(value);
     } else {
-      // TODO(40014): Remove cast when type promotion works.
-      // This would normally be `as T` but we use `as dynamic` to make the
-      // unneeded check be implicit to match dart2js unsound optimizations in
-      // the user code.
-      _future._completeWithValue(value as dynamic);
+      _future._completeWithValue(value);
     }
   }
 
@@ -220,7 +247,7 @@ class _AsyncAwaitCompleter<T> implements Completer<T> {
 /// Used as part of the runtime support for the async/await transformation.
 @pragma('dart2js:assumeDynamic') // Global type inference can't see call site.
 Completer<T> _makeAsyncAwaitCompleter<T>() {
-  return new _AsyncAwaitCompleter<T>();
+  return _AsyncAwaitCompleter<T>();
 }
 
 /// Initiates the computation of an `async` function and starts the body
@@ -232,8 +259,10 @@ Completer<T> _makeAsyncAwaitCompleter<T>() {
 /// Independently, it takes the [completer] and returns the future of the
 /// completer for convenience of the transformed code.
 dynamic _asyncStartSync(
-    _WrappedAsyncBody bodyFunction, _AsyncAwaitCompleter completer) {
-  bodyFunction(async_error_codes.SUCCESS, null);
+  _WrappedAsyncBody bodyFunction,
+  _AsyncAwaitCompleter completer,
+) {
+  bodyFunction(async_status_codes.SUCCESS, null);
   completer.isSync = true;
   return completer.future;
 }
@@ -243,8 +272,8 @@ dynamic _asyncStartSync(
 /// Used as part of the runtime support for the async/await transformation.
 ///
 /// Arranges for [bodyFunction] to be called when the future or value [object]
-/// is completed with a code [async_error_codes.SUCCESS] or
-/// [async_error_codes.ERROR] depending on the success of the future.
+/// is completed with a code [async_status_codes.SUCCESS] or
+/// [async_status_codes.ERROR] depending on the success of the future.
 dynamic _asyncAwait(dynamic object, _WrappedAsyncBody bodyFunction) {
   _awaitOnObject(object, bodyFunction);
 }
@@ -267,7 +296,9 @@ dynamic _asyncReturn(dynamic object, Completer completer) {
 dynamic _asyncRethrow(dynamic object, Completer completer) {
   // The error is a js-error.
   completer.completeError(
-      unwrapException(object), getTraceFromException(object));
+    unwrapException(object),
+    getTraceFromException(object),
+  );
 }
 
 /// Awaits on the given [object].
@@ -279,12 +310,14 @@ dynamic _asyncRethrow(dynamic object, Completer completer) {
 /// when the future completes.
 void _awaitOnObject(object, _WrappedAsyncBody bodyFunction) {
   FutureOr<dynamic> Function(dynamic) thenCallback =
-      (result) => bodyFunction(async_error_codes.SUCCESS, result);
+      (result) => bodyFunction(async_status_codes.SUCCESS, result);
 
   Function errorCallback = (dynamic error, StackTrace stackTrace) {
-    ExceptionAndStackTrace wrappedException =
-        new ExceptionAndStackTrace(error, stackTrace);
-    bodyFunction(async_error_codes.ERROR, wrappedException);
+    ExceptionAndStackTrace wrappedException = ExceptionAndStackTrace(
+      error,
+      stackTrace,
+    );
+    bodyFunction(async_status_codes.ERROR, wrappedException);
   };
 
   if (object is _Future) {
@@ -294,19 +327,19 @@ void _awaitOnObject(object, _WrappedAsyncBody bodyFunction) {
   } else if (object is Future) {
     object.then(thenCallback, onError: errorCallback);
   } else {
-    _Future future = new _Future().._setValue(object);
+    _Future future = _Future().._setValue(object);
     // We can skip the zone registration, since the bodyFunction is already
     // registered (see [_wrapJsFunctionForAsync]).
     future._thenAwait(thenCallback, errorCallback);
   }
 }
 
-typedef void _WrappedAsyncBody(int errorCode, dynamic result);
+typedef _WrappedAsyncBody = void Function(int errorCode, dynamic result);
 
 _WrappedAsyncBody _wrapJsFunctionForAsync(dynamic /* js function */ function) {
   var protected = JS(
-      '',
-      """
+    '',
+    """
         (function (fn, ERROR) {
           // Invokes [function] with [errorCode] and [result].
           //
@@ -324,8 +357,9 @@ _WrappedAsyncBody _wrapJsFunctionForAsync(dynamic /* js function */ function) {
             }
           }
         })(#, #)""",
-      function,
-      async_error_codes.ERROR);
+    function,
+    async_status_codes.ERROR,
+  );
 
   return Zone.current.registerBinaryCallback((int errorCode, dynamic result) {
     JS('', '#(#, #)', protected, errorCode, result);
@@ -338,11 +372,11 @@ _WrappedAsyncBody _wrapJsFunctionForAsync(dynamic /* js function */ function) {
 /// yield* and before starting the function.
 ///
 /// When the async* function wants to return it calls this function with
-/// [asyncBody] == [async_error_codes.SUCCESS], the asyncStarHelper takes this
+/// [asyncBody] == [async_status_codes.SUCCESS], the asyncStarHelper takes this
 /// as signal to close the stream.
 ///
 /// When the async* function wants to signal that an uncaught error was thrown,
-/// it calls this function with [asyncBody] == [async_error_codes.ERROR],
+/// it calls this function with [asyncBody] == [async_status_codes.ERROR],
 /// the streamHelper takes this as signal to addError [object] to the
 /// [controller] and close it.
 ///
@@ -351,7 +385,7 @@ _WrappedAsyncBody _wrapJsFunctionForAsync(dynamic /* js function */ function) {
 ///
 /// In the case of a yield or yield*, if the stream subscription has been
 /// canceled, schedules [asyncBody] to be called with
-/// [async_error_codes.STREAM_WAS_CANCELED].
+/// [async_status_codes.STREAM_WAS_CANCELED].
 ///
 /// If [object] is a single-yield [IterationMarker], adds the value of the
 /// [IterationMarker] to the stream. If the stream subscription has been
@@ -369,10 +403,11 @@ _WrappedAsyncBody _wrapJsFunctionForAsync(dynamic /* js function */ function) {
 /// If [object] is not a [Future], it is wrapped in a `Future.value`.
 /// The [asyncBody] is called on completion of the future (see [asyncHelper].
 void _asyncStarHelper(
-    dynamic object,
-    dynamic /* int | _WrappedAsyncBody */ bodyFunctionOrErrorCode,
-    _AsyncStarStreamController controller) {
-  if (identical(bodyFunctionOrErrorCode, async_error_codes.SUCCESS)) {
+  dynamic object,
+  dynamic /* int | _WrappedAsyncBody */ bodyFunctionOrErrorCode,
+  _AsyncStarStreamController controller,
+) {
+  if (identical(bodyFunctionOrErrorCode, async_status_codes.SUCCESS)) {
     // This happens on return from the async* function.
     if (controller.isCanceled) {
       controller.cancelationFuture!._completeWithValue(null);
@@ -380,14 +415,18 @@ void _asyncStarHelper(
       controller.close();
     }
     return;
-  } else if (identical(bodyFunctionOrErrorCode, async_error_codes.ERROR)) {
+  } else if (identical(bodyFunctionOrErrorCode, async_status_codes.ERROR)) {
     // The error is a js-error.
     if (controller.isCanceled) {
       controller.cancelationFuture!._completeError(
-          unwrapException(object), getTraceFromException(object));
+        unwrapException(object),
+        getTraceFromException(object),
+      );
     } else {
       controller.addError(
-          unwrapException(object), getTraceFromException(object));
+        unwrapException(object),
+        getTraceFromException(object),
+      );
       controller.close();
     }
     return;
@@ -396,7 +435,7 @@ void _asyncStarHelper(
   _WrappedAsyncBody bodyFunction = bodyFunctionOrErrorCode;
   if (object is _IterationMarker) {
     if (controller.isCanceled) {
-      bodyFunction(async_error_codes.STREAM_WAS_CANCELED, null);
+      bodyFunction(async_status_codes.STREAM_WAS_CANCELED, null);
       return;
     }
     if (object.state == _IterationMarker.YIELD_SINGLE) {
@@ -410,22 +449,26 @@ void _asyncStarHelper(
           controller.isSuspended = true;
           return;
         }
-        bodyFunction(async_error_codes.SUCCESS, null);
+        bodyFunction(
+          controller.isCanceled
+              ? async_status_codes.STREAM_WAS_CANCELED
+              : async_status_codes.SUCCESS,
+          null,
+        );
       });
       return;
     } else if (object.state == _IterationMarker.YIELD_STAR) {
       Stream stream = object.value;
       // Errors of [stream] are passed though to the main stream. (see
       // [AsyncStreamController.addStream]).
-      // TODO(sigurdm): The spec is not very clear here. Clarify with Gilad.
       controller.addStream(stream).then((_) {
-        // No check for isPaused here because the spec 17.16.2 only
-        // demands checks *before* each element in [stream] not after the last
-        // one. On the other hand we check for isCanceled, as that check happens
-        // after insertion of each element.
-        int errorCode = controller.isCanceled
-            ? async_error_codes.STREAM_WAS_CANCELED
-            : async_error_codes.SUCCESS;
+        // No need to check for pause because to get here the stream either
+        // completed normally or was cancelled. The stream cannot be paused
+        // after either of these states.
+        int errorCode =
+            controller.isCanceled
+                ? async_status_codes.STREAM_WAS_CANCELED
+                : async_status_codes.SUCCESS;
         bodyFunction(errorCode, null);
       });
       return;
@@ -481,34 +524,38 @@ class _AsyncStarStreamController<T> {
   _AsyncStarStreamController(_WrappedAsyncBody body) {
     _resumeBody() {
       scheduleMicrotask(() {
-        body(async_error_codes.SUCCESS, null);
+        body(async_status_codes.SUCCESS, null);
       });
     }
 
-    controller = new StreamController<T>(onListen: () {
-      _resumeBody();
-    }, onResume: () {
-      // Only schedule again if the async* function actually is suspended.
-      // Resume directly instead of scheduling, so that the sequence
-      // `pause-resume-pause` will result in one extra event produced.
-      if (isSuspended) {
-        isSuspended = false;
+    controller = StreamController<T>(
+      onListen: () {
         _resumeBody();
-      }
-    }, onCancel: () {
-      // If the async* is finished we ignore cancel events.
-      if (!controller.isClosed) {
-        cancelationFuture = new _Future();
+      },
+      onResume: () {
+        // Only schedule again if the async* function actually is suspended.
+        // Resume directly instead of scheduling, so that the sequence
+        // `pause-resume-pause` will result in one extra event produced.
         if (isSuspended) {
-          // Resume the suspended async* function to run finalizers.
           isSuspended = false;
-          scheduleMicrotask(() {
-            body(async_error_codes.STREAM_WAS_CANCELED, null);
-          });
+          _resumeBody();
         }
-        return cancelationFuture;
-      }
-    });
+      },
+      onCancel: () {
+        // If the async* is finished we ignore cancel events.
+        if (!controller.isClosed) {
+          cancelationFuture = _Future();
+          if (isSuspended) {
+            // Resume the suspended async* function to run finalizers.
+            isSuspended = false;
+            scheduleMicrotask(() {
+              body(async_status_codes.STREAM_WAS_CANCELED, null);
+            });
+          }
+          return cancelationFuture;
+        }
+      },
+    );
   }
 }
 
@@ -517,7 +564,7 @@ class _AsyncStarStreamController<T> {
 /// Used as part of the runtime support for the async/await transformation.
 @pragma('dart2js:assumeDynamic') // Global type inference can't see call site.
 _makeAsyncStarStreamController<T>(_WrappedAsyncBody body) {
-  return new _AsyncStarStreamController<T>(body);
+  return _AsyncStarStreamController<T>(body);
 }
 
 class _IterationMarker {
@@ -532,7 +579,7 @@ class _IterationMarker {
   const _IterationMarker._(this.state, this.value);
 
   static yieldStar(dynamic /* Iterable or Stream */ values) {
-    return new _IterationMarker._(YIELD_STAR, values);
+    return _IterationMarker._(YIELD_STAR, values);
   }
 
   static endOfIteration() {
@@ -540,11 +587,11 @@ class _IterationMarker {
   }
 
   static yieldSingle(dynamic value) {
-    return new _IterationMarker._(YIELD_SINGLE, value);
+    return _IterationMarker._(YIELD_SINGLE, value);
   }
 
   static uncaughtError(dynamic error) {
-    return new _IterationMarker._(UNCAUGHT_ERROR, error);
+    return _IterationMarker._(UNCAUGHT_ERROR, error);
   }
 
   toString() => "IterationMarker($state, $value)";
@@ -557,16 +604,19 @@ class _SyncStarIterator<T> implements Iterator<T> {
   // statements. For non-sync* iterators, [_nestedIterator] contains the
   // iterator. We delegate to [_nestedIterator] when it is not `null`.
   //
-  // For nested sync* iterators, [this] iterator acts on behalf of the innermost
+  // For nested sync* iterators, this [Iterator] acts on behalf of the innermost
   // nested sync* iterator. The current state machine is suspended on a stack
   // until the inner state machine ends.
 
   // The state machine for the innermost _SyncStarIterator.
-  dynamic _body;
+  Object? _body;
 
   // The current value, unless iterating a non-sync* nested iterator.
   T? _current = null;
 
+  // Value passed back from state machine for uncaught exceptions.
+  Object? _datum;
+
   // This is the nested iterator when iterating a yield* of a non-sync iterator.
   Iterator<T>? _nestedIterator = null;
 
@@ -577,100 +627,133 @@ class _SyncStarIterator<T> implements Iterator<T> {
   _SyncStarIterator(this._body);
 
   T get current {
-    var nested = _nestedIterator;
-    if (nested == null) return _current as dynamic; // implicit: as T;
-    return nested.current;
+    return _current as dynamic; // implicit: as T;
   }
 
-  _runBody() {
-    // TODO(sra): Find a way to hard-wire SUCCESS and ERROR codes.
-    return JS(
-        '',
-        '''
-        // Invokes [body] with [errorCode] and [result].
-        //
-        // If (and as long as) the invocation throws, calls [function] again,
-        // with an error-code.
-        (function(body, SUCCESS, ERROR) {
-          var errorValue, errorCode = SUCCESS;
-          while (true) {
-            try {
-              return body(errorCode, errorValue);
-            } catch (error) {
-              errorValue = error;
-              errorCode = ERROR;
-            }
-          }
-        })(#, #, #)''',
-        _body,
-        async_error_codes.SUCCESS,
-        async_error_codes.ERROR);
+  _resumeBody(int errorCode, Object? errorValue) {
+    final body = _body;
+    while (true) {
+      try {
+        return JS('', '#(#, #, #)', body, this, errorCode, errorValue);
+      } catch (error) {
+        errorValue = JS_RAW_EXCEPTION();
+        errorCode = async_status_codes.ERROR;
+      }
+    }
   }
 
   bool moveNext() {
+    if (JS_FALSE()) _modelGeneratedCode();
+    Object? errorValue;
+    int errorCode = async_status_codes.SUCCESS;
     while (true) {
-      if (_nestedIterator != null) {
-        if (_nestedIterator!.moveNext()) {
-          return true;
-        } else {
+      final nestedIterator = _nestedIterator;
+      if (nestedIterator != null) {
+        try {
+          if (nestedIterator.moveNext()) {
+            _current = nestedIterator.current;
+            return true;
+          } else {
+            _nestedIterator = null;
+          }
+        } catch (error) {
+          errorValue = JS_RAW_EXCEPTION();
+          errorCode = async_status_codes.ERROR;
           _nestedIterator = null;
         }
       }
-      var value = _runBody();
-      if (value is _IterationMarker) {
-        int state = value.state;
-        if (state == _IterationMarker.ITERATION_ENDED) {
-          var suspendedBodies = _suspendedBodies;
-          if (suspendedBodies == null || suspendedBodies.isEmpty) {
-            _current = null;
-            // Rely on [_body] to repeatedly return `ITERATION_ENDED`.
-            return false;
-          }
-          // Resume the innermost suspended iterator.
-          _body = suspendedBodies.removeLast();
-          continue;
-        } else if (state == _IterationMarker.UNCAUGHT_ERROR) {
-          // Rely on [_body] to repeatedly return `UNCAUGHT_ERROR`.
-          // This is a wrapped exception, so we use JavaScript throw to throw
-          // it.
-          JS('', 'throw #', value.value);
-        } else {
-          assert(state == _IterationMarker.YIELD_STAR);
-          // Casting to avoid the `.iterator` dynamic call would impact the
-          // performance of yield*.
-          // ignore: avoid_dynamic_calls
-          Iterator<T> inner = value.value.iterator;
-          if (inner is _SyncStarIterator) {
-            // The test needs to be 'is _SyncStarIterator<T>' for promotion to
-            // work. However, that test is much more expensive, so we use an
-            // unsafe cast.
-            _SyncStarIterator<T> innerSyncStarIterator = JS('', '#', inner);
-            // Suspend the current state machine and start acting on behalf of
-            // the nested state machine.
-            //
-            // TODO(sra): Recognize "tail yield*" statements and avoid
-            // suspending the current body when all it will do is step without
-            // effect to ITERATION_ENDED.
-            (_suspendedBodies ??= []).add(_body);
-            _body = innerSyncStarIterator._body;
-            continue;
-          } else {
-            _nestedIterator = inner;
-            // TODO(32956): Change to the following when strong-mode is the only
-            // option:
-            //
-            //     _nestedIterator = JS<Iterator<T>>('','#', inner);
-            continue;
-          }
-        }
-      } else {
-        // TODO(32956): Remove this test.
-        _current = JS<T>('', '#', value);
+
+      var value = _resumeBody(errorCode, errorValue);
+
+      if (async_status_codes.SYNC_STAR_YIELD == value) {
+        // The state-machine has assgned the value to _current.
         return true;
       }
+      if (async_status_codes.SYNC_STAR_DONE == value) {
+        _current = null;
+        var suspendedBodies = _suspendedBodies;
+        if (suspendedBodies == null || suspendedBodies.isEmpty) {
+          // Overwrite the body with a stub for an empty iterable. If [moveNext]
+          // is called 'too many' times, it continues to return `false`.
+          _body = RAW_DART_FUNCTION_REF(_terminatedBody);
+          return false;
+        }
+        // Resume the innermost suspended iterator.
+        _body = suspendedBodies.removeLast();
+        errorCode = async_status_codes.SUCCESS;
+        errorValue = null;
+        continue;
+      }
+      if (async_status_codes.SYNC_STAR_YIELD_STAR == value) {
+        // The call to _yieldStar has modified the state.
+        errorCode = async_status_codes.SUCCESS;
+        errorValue = null;
+        continue;
+      }
+      if (async_status_codes.SYNC_STAR_UNCAUGHT_EXCEPTION == value) {
+        errorValue = _datum;
+        _datum = null;
+        var suspendedBodies = _suspendedBodies;
+        if (suspendedBodies == null || suspendedBodies.isEmpty) {
+          _current = null;
+          // Overwrite the body with a stub for an empty iterable. If [moveNext]
+          // is called after the exception propagates out of the `yield*` stack,
+          // it will return `false`.
+          _body = RAW_DART_FUNCTION_REF(_terminatedBody);
+
+          // This is a wrapped exception, so we use JavaScript throw to throw
+          // it.
+          JS('', 'throw #', errorValue);
+          // The above is not seen as terminating, so we need this return:
+          return false; // unreachable
+        }
+        // Resume the innermost suspended iterator.
+        _body = suspendedBodies.removeLast();
+        errorCode = async_status_codes.ERROR;
+        continue;
+      }
+      throw StateError('sync*');
     }
-    return false; // TODO(sra): Fix type inference so that this is not needed.
+    // TODO(http://dartbug.com/52166): Fix type inference so that this return
+    // statement is not needed.
+    return false;
   }
+
+  static _terminatedBody(_1, _2, _3) => async_status_codes.SYNC_STAR_DONE;
+
+  // Called from generated code.
+  @pragma('dart2js:parameter:trust')
+  int _yieldStar(Iterable<T> iterable) {
+    if (iterable is _SyncStarIterable) {
+      // Promotion fails, so we need this zero-cost 'cast'.
+      _SyncStarIterable<T> syncStarIterable = JS('', '#', iterable);
+      _SyncStarIterator inner = syncStarIterable.iterator;
+      // Suspend the current state machine and start acting on behalf of
+      // the nested state machine.
+      //
+      // TODO(sra): Recognize "tail yield*" statements and avoid
+      // suspending the current body when all it will do is step without
+      // effect to ITERATION_ENDED.
+      (_suspendedBodies ??= []).add(_body);
+      _body = inner._body;
+      return async_status_codes.SYNC_STAR_YIELD_STAR;
+    } else {
+      _nestedIterator = iterable.iterator;
+      return async_status_codes.SYNC_STAR_YIELD_STAR;
+    }
+  }
+
+  // This code is assumed present by global type inference and models the types
+  // flowing to various fields and methods from code inserted by the
+  // state-machine rewriter.
+  _modelGeneratedCode() {
+    _yieldStar(_confuse<Iterable<T>>(this));
+    _current = _confuse<T?>(null);
+    _datum = _confuse<Object>(this);
+  }
+
+  @pragma('dart2js:assumeDynamic')
+  static T _confuse<T>(dynamic x) => x;
 }
 
 /// Creates an Iterable for a `sync*` function.
@@ -678,7 +761,7 @@ class _SyncStarIterator<T> implements Iterator<T> {
 /// Used as part of the runtime support for the async/await transformation.
 @pragma('dart2js:assumeDynamic') // Global type inference can't see call site.
 _SyncStarIterable<T> _makeSyncStarIterable<T>(body) {
-  return new _SyncStarIterable<T>(body);
+  return _SyncStarIterable<T>(body);
 }
 
 /// An Iterable corresponding to a sync* method.
@@ -693,6 +776,20 @@ class _SyncStarIterable<T> extends Iterable<T> {
 
   _SyncStarIterable(this._outerHelper);
 
-  Iterator<T> get iterator =>
-      new _SyncStarIterator<T>(JS('', '#()', _outerHelper));
+  @pragma('dart2js:prefer-inline')
+  _SyncStarIterator<T> get iterator =>
+      _SyncStarIterator<T>(JS('', '#()', _outerHelper));
 }
+
+/// Wraps an `await`ed expression in [Future.value] if needed.
+///
+/// If an expression `e` has a static type of `S`, then `await e` must first
+/// check if the runtime type of `e` is `Future<flatten(S)>`. If it is, `e` can
+/// be `await`ed directly. Otherwise, we must `await Future.value(e)`. Here, [T]
+/// is expected to be `flatten(S)`.
+///
+/// It suffices to use `_Future.value` rather than `Future.value` - see the
+/// comments on https://github.com/dart-lang/sdk/issues/50601.
+@pragma('dart2js:prefer-inline')
+Future<T> _wrapAwaitedExpression<T>(Object? e) =>
+    e is Future<T> ? e : _Future<T>.value(unsafeCast<T>(e));
