@@ -10,14 +10,18 @@ const int _maxUtf16 = 0xffff;
 const int _maxUnicode = 0x10ffff;
 
 @patch
+@pragma('vm:deeply-immutable')
 @pragma("vm:entry-point")
 class String {
   @patch
-  factory String.fromCharCodes(Iterable<int> charCodes,
-      [int start = 0, int? end]) {
+  factory String.fromCharCodes(
+    Iterable<int> charCodes, [
+    int start = 0,
+    int? end,
+  ]) {
     // TODO: Remove these null checks once all code is opted into strong nonnullable mode.
-    if (charCodes == null) throw new ArgumentError.notNull("charCodes");
-    if (start == null) throw new ArgumentError.notNull("start");
+    if (charCodes == null) throw ArgumentError.notNull("charCodes");
+    if (start == null) throw ArgumentError.notNull("start");
     return _StringBase.createFromCharCodes(charCodes, start, end, null);
   }
 
@@ -28,28 +32,26 @@ class String {
         return _OneByteString._allocate(1).._setAt(0, charCode);
       }
       if (charCode <= 0xffff) {
-        return _StringBase._createFromCodePoints(
-            new _List(1)..[0] = charCode, 0, 1);
+        return _TwoByteString._allocate(1).._setAt(0, charCode);
       }
       if (charCode <= 0x10ffff) {
         var low = 0xDC00 | (charCode & 0x3ff);
         int bits = charCode - 0x10000;
         var high = 0xD800 | (bits >> 10);
-        return _StringBase._createFromCodePoints(
-            new _List(2)
-              ..[0] = high
-              ..[1] = low,
-            0,
-            2);
+        return _TwoByteString._allocate(2)
+          .._setAt(0, high)
+          .._setAt(1, low);
       }
     }
-    throw new RangeError.range(charCode, 0, 0x10ffff);
+    throw RangeError.range(charCode, 0, 0x10ffff);
   }
 
   @patch
   @pragma("vm:external-name", "String_fromEnvironment")
-  external const factory String.fromEnvironment(String name,
-      {String defaultValue = ""});
+  external const factory String.fromEnvironment(
+    String name, {
+    String defaultValue = "",
+  });
 
   bool get _isOneByte;
   String _substringUnchecked(int startIndex, int endIndex);
@@ -59,6 +61,7 @@ class String {
  * [_StringBase] contains common methods used by concrete String
  * implementations, e.g., _OneByteString.
  */
+@pragma('vm:deeply-immutable')
 abstract final class _StringBase implements String {
   bool _isWhitespace(int codeUnit);
 
@@ -93,7 +96,7 @@ abstract final class _StringBase implements String {
   static const int _maxJoinReplaceOneByteStringLength = 500;
 
   factory _StringBase._uninstantiable() {
-    throw new UnsupportedError("_StringBase can't be instantiated");
+    throw UnsupportedError("_StringBase can't be instantiated");
   }
 
   @pragma("vm:recognized", "asm-intrinsic")
@@ -109,8 +112,7 @@ abstract final class _StringBase implements String {
   bool get _isOneByte {
     // Alternatively return false and override it on one-byte string classes.
     int id = ClassID.getID(this);
-    return id == ClassID.cidOneByteString ||
-        id == ClassID.cidExternalOneByteString;
+    return id == ClassID.cidOneByteString;
   }
 
   /**
@@ -124,109 +126,137 @@ abstract final class _StringBase implements String {
    * It's `null` if unknown.
    */
   static String createFromCharCodes(
-      Iterable<int> charCodes, int start, int? end, int? limit) {
+    Iterable<int> charCodes,
+    int start,
+    int? end,
+    int? limit,
+  ) {
+    // Validate start/end first.
+    RangeError.checkNotNegative(start, "start");
+    if (end != null) {
+      var maxLength = end - start;
+      if (maxLength < 0) {
+        throw RangeError.range(end, start, null, "end");
+      }
+      if (maxLength == 0) {
+        return "";
+      }
+    }
     // TODO(srdjan): Also skip copying of wide typed arrays.
     final ccid = ClassID.getID(charCodes);
     if ((ccid != ClassID.cidArray) &&
         (ccid != ClassID.cidGrowableObjectArray) &&
         (ccid != ClassID.cidImmutableArray)) {
       if (charCodes is Uint8List) {
-        final actualEnd =
-            RangeError.checkValidRange(start, end, charCodes.length);
-        return _createOneByteString(charCodes, start, actualEnd - start);
+        final int codeCount = charCodes.length;
+        if (codeCount <= start) return "";
+        if (end == null || end > codeCount) {
+          end = codeCount;
+        }
+        return _createOneByteString(charCodes, start, end - start);
       } else if (charCodes is! Uint16List) {
         return _createStringFromIterable(charCodes, start, end);
       }
     }
     final int codeCount = charCodes.length;
-    final actualEnd = RangeError.checkValidRange(start, end, codeCount);
-    final len = actualEnd - start;
+    if (codeCount <= start) return "";
+    if (end == null || end > codeCount) {
+      end = codeCount;
+    }
+    final len = end - start;
     if (len == 0) return "";
 
     final typedCharCodes = unsafeCast<List<int>>(charCodes);
 
-    final int actualLimit =
-        limit ?? _scanCodeUnits(typedCharCodes, start, actualEnd);
+    final int actualLimit = limit ?? _scanCodeUnits(typedCharCodes, start, end);
     if (actualLimit < 0) {
-      throw new ArgumentError(typedCharCodes);
+      throw ArgumentError(typedCharCodes);
     }
     if (actualLimit <= _maxLatin1) {
       return _createOneByteString(typedCharCodes, start, len);
     }
     if (actualLimit <= _maxUtf16) {
       return _TwoByteString._allocateFromTwoByteList(
-          typedCharCodes, start, actualEnd);
+        typedCharCodes,
+        start,
+        end,
+      );
     }
     // TODO(lrn): Consider passing limit to _createFromCodePoints, because
     // the function is currently fully generic and doesn't know that its
-    // charCodes are not all Latin-1 or Utf-16.
-    return _createFromCodePoints(typedCharCodes, start, actualEnd);
+    // charCodes are not all Latin-1 or UTF-16.
+    return _createFromCodePoints(typedCharCodes, start, end);
   }
 
   static int _scanCodeUnits(List<int> charCodes, int start, int end) {
     int bits = 0;
     for (int i = start; i < end; i++) {
       int code = charCodes[i];
-      if (code is! _Smi) throw new ArgumentError(charCodes);
+      if (code is! _Smi) throw ArgumentError(charCodes);
       bits |= code;
     }
     return bits;
   }
 
   static String _createStringFromIterable(
-      Iterable<int> charCodes, int start, int? end) {
+    Iterable<int> charCodes,
+    int start,
+    int? end,
+  ) {
+    assert(start >= 0);
+    assert(end == null || start <= end);
     // Treat charCodes as Iterable.
     if (charCodes is EfficientLengthIterable) {
-      int length = charCodes.length;
-      final endVal = RangeError.checkValidRange(start, end, length);
-      final charCodeList = new List<int>.from(
-          charCodes.take(endVal).skip(start),
-          growable: false);
-      return createFromCharCodes(charCodeList, 0, charCodeList.length, null);
+      final int codeCount = charCodes.length;
+      if (start >= codeCount) return "";
+      if (end == null || end > codeCount) {
+        end = codeCount;
+      }
     }
-    // Don't know length of iterable, so iterate and see if all the values
-    // are there.
-    if (start < 0) throw new RangeError.range(start, 0, charCodes.length);
     var it = charCodes.iterator;
     for (int i = 0; i < start; i++) {
-      if (!it.moveNext()) {
-        throw new RangeError.range(start, 0, i);
-      }
+      if (!it.moveNext()) return "";
     }
-    List<int> charCodeList;
-    int bits = 0; // Bitwise-or of all char codes in list.
-    final endVal = end;
-    if (endVal == null) {
-      var list = <int>[];
-      while (it.moveNext()) {
-        int code = it.current;
+    // Bitwise-or of all char codes in list.
+    // There are two valid ranges:
+    // 0x00-0xFF: Valid one-byte string.
+    // 0x100-0xFFFFF: Valid two-byte string with bits-values in range
+    //    0x10000-0xFFFFF already encoded as surrogate pairs.
+    // Numbers above that, or negative, correspond to input "char codes"
+    // outside of the range U+0000 .. U+10FFFF.
+    int bits = 0;
+    int takeCount = (end == null) ? -1 : (end - start); // -1 means no limit.
+    var list = <int>[];
+    while (takeCount != 0 && it.moveNext()) {
+      takeCount -= 1;
+      int code = it.current;
+      if (code <= 0xFFFF) {
         bits |= code;
         list.add(code);
+      } else {
+        code -= 0x10000;
+        // Any value in the range 0x100..0xFFFFF can be used for the second
+        // number. Using 0xD800 to represent containing surrogate pairs.
+        bits |= code | 0xD800;
+        list
+          ..add(0xD800 + (code >>> 10))
+          ..add(0xDC00 + (code & 0x3FF));
       }
-      charCodeList = makeListFixedLength<int>(list);
-    } else {
-      if (endVal < start) {
-        throw new RangeError.range(endVal, start, charCodes.length);
-      }
-      int len = endVal - start;
-      charCodeList = new List<int>.generate(len, (int i) {
-        if (!it.moveNext()) {
-          throw new RangeError.range(endVal, start, start + i);
-        }
-        int code = it.current;
-        bits |= code;
-        return code;
-      });
     }
-    int length = charCodeList.length;
-    if (bits < 0) {
-      throw new ArgumentError(charCodes);
+    if (bits < 0 || bits > 0xFFFFF) {
+      throw ArgumentError.value(
+        charCodes,
+        "charCodes",
+        "Contains invalid character code, not 0 <= code <= 0x10FFFF",
+      );
     }
+    List<int> codeUnitList = makeListFixedLength<int>(list);
+    int length = codeUnitList.length;
     bool isOneByteString = (bits <= _maxLatin1);
     if (isOneByteString) {
-      return _createOneByteString(charCodeList, 0, length);
+      return _createOneByteString(codeUnitList, 0, length);
     }
-    return createFromCharCodes(charCodeList, 0, length, bits);
+    return _TwoByteString._allocateFromTwoByteList(codeUnitList, 0, length);
   }
 
   // Inlining is disabled as a workaround to http://dartbug.com/37800.
@@ -251,13 +281,20 @@ abstract final class _StringBase implements String {
 
   @pragma("vm:external-name", "StringBase_createFromCodePoints")
   external static String _createFromCodePoints(
-      List<int> codePoints, int start, int end);
+    List<int> codePoints,
+    int start,
+    int end,
+  );
 
   @pragma("vm:recognized", "asm-intrinsic")
   @pragma("vm:external-name", "String_charAt")
   external String operator [](int index);
 
-  int codeUnitAt(int index); // Implemented in the subclasses.
+  @pragma("vm:recognized", "other")
+  @pragma("vm:prefer-inline")
+  @pragma("vm:idempotent")
+  @pragma("vm:exact-result-type", "dart:core#_Smi")
+  external int codeUnitAt(int index);
 
   @pragma("vm:recognized", "graph-intrinsic")
   @pragma("vm:exact-result-type", "dart:core#_Smi")
@@ -338,7 +375,7 @@ abstract final class _StringBase implements String {
 
   bool startsWith(Pattern pattern, [int index = 0]) {
     if ((index < 0) || (index > this.length)) {
-      throw new RangeError.range(index, 0, this.length);
+      throw RangeError.range(index, 0, this.length);
     }
     if (pattern is String) {
       return _substringMatches(index, pattern);
@@ -348,12 +385,21 @@ abstract final class _StringBase implements String {
 
   int indexOf(Pattern pattern, [int start = 0]) {
     if ((start < 0) || (start > this.length)) {
-      throw new RangeError.range(start, 0, this.length, "start");
+      throw RangeError.range(start, 0, this.length, "start");
     }
     if (pattern is String) {
       String other = pattern;
-      int maxIndex = this.length - other.length;
-      // TODO: Use an efficient string search (e.g. BMH).
+      var otherLength = other.length;
+      if (otherLength == 0) return start;
+      if (otherLength == 1) {
+        int codeUnit = other.codeUnitAt(0);
+        for (var index = start; index < this.length; index++) {
+          if (codeUnit == this.codeUnitAt(index)) return index;
+        }
+        return -1;
+      }
+      int maxIndex = this.length - otherLength;
+      // TODO: Use an efficient string search (e.g., BMH).
       for (int index = start; index <= maxIndex; index++) {
         if (_substringMatches(index, other)) {
           return index;
@@ -362,8 +408,6 @@ abstract final class _StringBase implements String {
       return -1;
     }
     for (int i = start; i <= this.length; i++) {
-      // TODO(11276); This has quadratic behavior because matchAsPrefix tries
-      // to find a later match too. Optimize matchAsPrefix to avoid this.
       if (pattern.matchAsPrefix(this, i) != null) return i;
     }
     return -1;
@@ -373,7 +417,7 @@ abstract final class _StringBase implements String {
     if (start == null) {
       start = this.length;
     } else if (start < 0 || start > this.length) {
-      throw new RangeError.range(start, 0, this.length);
+      throw RangeError.range(start, 0, this.length);
     }
     if (pattern is String) {
       String other = pattern;
@@ -387,8 +431,6 @@ abstract final class _StringBase implements String {
       return -1;
     }
     for (int i = start; i >= 0; i--) {
-      // TODO(11276); This has quadratic behavior because matchAsPrefix tries
-      // to find a later match too. Optimize matchAsPrefix to avoid this.
       if (pattern.matchAsPrefix(this, i) != null) return i;
     }
     return -1;
@@ -538,7 +580,7 @@ abstract final class _StringBase implements String {
   String operator *(int times) {
     if (times <= 0) return "";
     if (times == 1) return this;
-    StringBuffer buffer = new StringBuffer(this);
+    StringBuffer buffer = StringBuffer(this);
     for (int i = 1; i < times; i++) {
       buffer.write(this);
     }
@@ -548,7 +590,7 @@ abstract final class _StringBase implements String {
   String padLeft(int width, [String padding = ' ']) {
     int delta = width - this.length;
     if (delta <= 0) return this;
-    StringBuffer buffer = new StringBuffer();
+    StringBuffer buffer = StringBuffer();
     for (int i = 0; i < delta; i++) {
       buffer.write(padding);
     }
@@ -559,7 +601,7 @@ abstract final class _StringBase implements String {
   String padRight(int width, [String padding = ' ']) {
     int delta = width - this.length;
     if (delta <= 0) return this;
-    StringBuffer buffer = new StringBuffer(this);
+    StringBuffer buffer = StringBuffer(this);
     for (int i = 0; i < delta; i++) {
       buffer.write(padding);
     }
@@ -567,31 +609,22 @@ abstract final class _StringBase implements String {
   }
 
   bool contains(Pattern pattern, [int startIndex = 0]) {
-    if (pattern is String) {
-      if (startIndex < 0 || startIndex > this.length) {
-        throw new RangeError.range(startIndex, 0, this.length);
-      }
+    if (startIndex >= 0 && startIndex <= this.length) {
       return indexOf(pattern, startIndex) >= 0;
     }
-    return pattern.allMatches(this.substring(startIndex)).isNotEmpty;
+    throw RangeError.range(startIndex, 0, this.length, "startIndex");
   }
 
-  String replaceFirst(Pattern pattern, String replacement,
-      [int startIndex = 0]) {
-    // TODO: Remove these null checks once all code is opted into strong nonnullable mode.
-    if (pattern == null) {
-      throw new ArgumentError.notNull("pattern");
-    }
-    if (replacement == null) {
-      throw new ArgumentError.notNull("replacement");
-    }
-    if (startIndex == null) {
-      throw new ArgumentError.notNull("startIndex");
-    }
+  String replaceFirst(
+    Pattern pattern,
+    String replacement, [
+    int startIndex = 0,
+  ]) {
     RangeError.checkValueInInterval(startIndex, 0, this.length, "startIndex");
-    Iterator iterator = startIndex == 0
-        ? pattern.allMatches(this).iterator
-        : pattern.allMatches(this, startIndex).iterator;
+    Iterator iterator =
+        startIndex == 0
+            ? pattern.allMatches(this).iterator
+            : pattern.allMatches(this, startIndex).iterator;
     if (!iterator.moveNext()) return this;
     Match match = iterator.current;
     return replaceRange(match.start, match.end, replacement);
@@ -617,7 +650,11 @@ abstract final class _StringBase implements String {
     if (replacement.length > 0) slices.add(replacement);
     _addReplaceSlice(slices, localEnd, length);
     return _joinReplaceAllResult(
-        this, slices, totalLength, replacementIsOneByte);
+      this,
+      slices,
+      totalLength,
+      replacementIsOneByte,
+    );
   }
 
   static int _addReplaceSlice(List matches, int start, int end) {
@@ -634,8 +671,8 @@ abstract final class _StringBase implements String {
   }
 
   String replaceAll(Pattern pattern, String replacement) {
-    if (pattern == null) throw new ArgumentError.notNull("pattern");
-    if (replacement == null) throw new ArgumentError.notNull("replacement");
+    if (pattern == null) throw ArgumentError.notNull("pattern");
+    if (replacement == null) throw ArgumentError.notNull("replacement");
 
     int startIndex = 0;
     // String fragments that replace the prefix [this] up to [startIndex].
@@ -663,7 +700,7 @@ abstract final class _StringBase implements String {
     if (replacementIsOneByte &&
         length < _maxJoinReplaceOneByteStringLength &&
         this._isOneByte) {
-      // TODO(lrn): Is there a cut-off point, or is runtime always faster?
+      // TODO: Is there a cut-off point, or is runtime always faster?
       return _joinReplaceAllOneByteResult(this, matches, length);
     }
     return _joinReplaceAllResult(this, matches, length, replacementIsOneByte);
@@ -674,7 +711,10 @@ abstract final class _StringBase implements String {
    * is always a [_OneByteString].
    */
   static String _joinReplaceAllOneByteResult(
-      String base, List matches, int length) {
+    String base,
+    List matches,
+    int length,
+  ) {
     _OneByteString result = _OneByteString._allocate(length);
     int writeIndex = 0;
     for (int i = 0; i < matches.length; i++) {
@@ -723,11 +763,15 @@ abstract final class _StringBase implements String {
    */
   @pragma("vm:external-name", "StringBase_joinReplaceAllResult")
   external static String _joinReplaceAllResult(
-      String base, List matches, int length, bool replacementStringsAreOneByte);
+    String base,
+    List matches,
+    int length,
+    bool replacementStringsAreOneByte,
+  );
 
   String replaceAllMapped(Pattern pattern, String replace(Match match)) {
-    if (pattern == null) throw new ArgumentError.notNull("pattern");
-    if (replace == null) throw new ArgumentError.notNull("replace");
+    if (pattern == null) throw ArgumentError.notNull("pattern");
+    if (replace == null) throw ArgumentError.notNull("replace");
     List matches = [];
     int length = 0;
     int startIndex = 0;
@@ -749,14 +793,21 @@ abstract final class _StringBase implements String {
       return _joinReplaceAllOneByteResult(this, matches, length);
     }
     return _joinReplaceAllResult(
-        this, matches, length, replacementStringsAreOneByte);
+      this,
+      matches,
+      length,
+      replacementStringsAreOneByte,
+    );
   }
 
-  String replaceFirstMapped(Pattern pattern, String replace(Match match),
-      [int startIndex = 0]) {
-    if (pattern == null) throw new ArgumentError.notNull("pattern");
-    if (replace == null) throw new ArgumentError.notNull("replace");
-    if (startIndex == null) throw new ArgumentError.notNull("startIndex");
+  String replaceFirstMapped(
+    Pattern pattern,
+    String replace(Match match), [
+    int startIndex = 0,
+  ]) {
+    if (pattern == null) throw ArgumentError.notNull("pattern");
+    if (replace == null) throw ArgumentError.notNull("replace");
+    if (startIndex == null) throw ArgumentError.notNull("startIndex");
     RangeError.checkValueInInterval(startIndex, 0, this.length, "startIndex");
 
     var matches = pattern.allMatches(this, startIndex).iterator;
@@ -770,14 +821,16 @@ abstract final class _StringBase implements String {
   static String _stringIdentity(String string) => string;
 
   String _splitMapJoinEmptyString(
-      String onMatch(Match match), String onNonMatch(String nonMatch)) {
+    String onMatch(Match match),
+    String onNonMatch(String nonMatch),
+  ) {
     // Pattern is the empty string.
-    StringBuffer buffer = new StringBuffer();
+    StringBuffer buffer = StringBuffer();
     int length = this.length;
     int i = 0;
     buffer.write(onNonMatch(""));
     while (i < length) {
-      buffer.write(onMatch(new _StringMatch(i, this, "")));
+      buffer.write(onMatch(_StringMatch(i, this, "")));
       // Special case to avoid splitting a surrogate pair.
       int code = this.codeUnitAt(i);
       if ((code & ~0x3FF) == 0xD800 && length > i + 1) {
@@ -793,15 +846,18 @@ abstract final class _StringBase implements String {
       buffer.write(onNonMatch(this[i]));
       i++;
     }
-    buffer.write(onMatch(new _StringMatch(i, this, "")));
+    buffer.write(onMatch(_StringMatch(i, this, "")));
     buffer.write(onNonMatch(""));
     return buffer.toString();
   }
 
-  String splitMapJoin(Pattern pattern,
-      {String onMatch(Match match)?, String onNonMatch(String nonMatch)?}) {
+  String splitMapJoin(
+    Pattern pattern, {
+    String onMatch(Match match)?,
+    String onNonMatch(String nonMatch)?,
+  }) {
     if (pattern == null) {
-      throw new ArgumentError.notNull("pattern");
+      throw ArgumentError.notNull("pattern");
     }
     onMatch ??= _matchString;
     onNonMatch ??= _stringIdentity;
@@ -811,7 +867,7 @@ abstract final class _StringBase implements String {
         return _splitMapJoinEmptyString(onMatch, onNonMatch);
       }
     }
-    StringBuffer buffer = new StringBuffer();
+    StringBuffer buffer = StringBuffer();
     int startIndex = 0;
     for (Match match in pattern.allMatches(this)) {
       buffer.write(onNonMatch(this.substring(startIndex, match.start)));
@@ -876,20 +932,19 @@ abstract final class _StringBase implements String {
 
   static ArgumentError _interpolationError(Object? o, Object? result) {
     // Since Dart 2.0, [result] can only be null.
-    return new ArgumentError.value(
-        o, "object", "toString method returned 'null'");
+    return ArgumentError.value(o, "object", "toString method returned 'null'");
   }
 
   Iterable<Match> allMatches(String string, [int start = 0]) {
     if (start < 0 || start > string.length) {
-      throw new RangeError.range(start, 0, string.length, "start");
+      throw RangeError.range(start, 0, string.length, "start");
     }
-    return new _StringAllMatchesIterable(string, this, start);
+    return _StringAllMatchesIterable(string, this, start);
   }
 
   Match? matchAsPrefix(String string, [int start = 0]) {
     if (start < 0 || start > string.length) {
-      throw new RangeError.range(start, 0, string.length);
+      throw RangeError.range(start, 0, string.length);
     }
     if (start + this.length > string.length) return null;
     for (int i = 0; i < this.length; i++) {
@@ -897,13 +952,15 @@ abstract final class _StringBase implements String {
         return null;
       }
     }
-    return new _StringMatch(start, string, this);
+    return _StringMatch(start, string, this);
   }
 
   List<String> split(Pattern pattern) {
     if ((pattern is String) && pattern.isEmpty) {
-      List<String> result =
-          new List<String>.generate(this.length, (int i) => this[i]);
+      List<String> result = List<String>.generate(
+        this.length,
+        (int i) => this[i],
+      );
       return result;
     }
     int length = this.length;
@@ -938,9 +995,9 @@ abstract final class _StringBase implements String {
     return result;
   }
 
-  List<int> get codeUnits => new CodeUnits(this);
+  List<int> get codeUnits => CodeUnits(this);
 
-  Runes get runes => new Runes(this);
+  Runes get runes => Runes(this);
 
   @pragma("vm:external-name", "String_toUpperCase")
   external String toUpperCase();
@@ -982,6 +1039,7 @@ int _clampedPositiveProduct(int a, int b) {
   return product;
 }
 
+@pragma('vm:deeply-immutable')
 @pragma("vm:entry-point")
 final class _OneByteString extends _StringBase {
   factory _OneByteString._uninstantiable() {
@@ -993,11 +1051,6 @@ final class _OneByteString extends _StringBase {
   @pragma("vm:external-name", "String_getHashCode")
   external int get hashCode;
 
-  @pragma("vm:recognized", "graph-intrinsic")
-  @pragma("vm:exact-result-type", "dart:core#_Smi")
-  @pragma("vm:external-name", "String_codeUnitAt")
-  external int codeUnitAt(int index);
-
   bool _isWhitespace(int codeUnit) {
     return _StringBase._isOneByteWhitespace(codeUnit);
   }
@@ -1066,8 +1119,7 @@ final class _OneByteString extends _StringBase {
     // Specialize for single character pattern.
     final pCid = ClassID.getID(pattern);
     if ((pCid == ClassID.cidOneByteString) ||
-        (pCid == ClassID.cidTwoByteString) ||
-        (pCid == ClassID.cidExternalOneByteString)) {
+        (pCid == ClassID.cidTwoByteString)) {
       final String patternAsString = unsafeCast<String>(pattern);
       final len = this.length;
       if ((patternAsString.length == 1) && (start >= 0) && (start < len)) {
@@ -1089,8 +1141,7 @@ final class _OneByteString extends _StringBase {
   bool contains(Pattern pattern, [int start = 0]) {
     final pCid = ClassID.getID(pattern);
     if ((pCid == ClassID.cidOneByteString) ||
-        (pCid == ClassID.cidTwoByteString) ||
-        (pCid == ClassID.cidExternalOneByteString)) {
+        (pCid == ClassID.cidTwoByteString)) {
       final String patternAsString = unsafeCast<String>(pattern);
       final len = this.length;
       if ((patternAsString.length == 1) && (start >= 0) && (start < len)) {
@@ -1129,8 +1180,7 @@ final class _OneByteString extends _StringBase {
 
   String padLeft(int width, [String padding = ' ']) {
     int padCid = ClassID.getID(padding);
-    if ((padCid != ClassID.cidOneByteString) &&
-        (padCid != ClassID.cidExternalOneByteString)) {
+    if (padCid != ClassID.cidOneByteString) {
       return super.padLeft(width, padding);
     }
     int length = this.length;
@@ -1160,8 +1210,7 @@ final class _OneByteString extends _StringBase {
 
   String padRight(int width, [String padding = ' ']) {
     int padCid = ClassID.getID(padding);
-    if ((padCid != ClassID.cidOneByteString) &&
-        (padCid != ClassID.cidExternalOneByteString)) {
+    if (padCid != ClassID.cidOneByteString) {
       return super.padRight(width, padding);
     }
     int length = this.length;
@@ -1293,7 +1342,10 @@ final class _OneByteString extends _StringBase {
 
   @pragma("vm:external-name", "OneByteString_allocateFromOneByteList")
   external static _OneByteString _allocateFromOneByteList(
-      List<int> list, int start, int end);
+    List<int> list,
+    int start,
+    int end,
+  );
 
   // This is internal helper method. Code point value must be a valid
   // Latin1 value (0..0xFF), index must be valid.
@@ -1303,7 +1355,7 @@ final class _OneByteString extends _StringBase {
   }
 
   // Should be optimizable to a memory move.
-  // Accepts both _OneByteString and _ExternalOneByteString as argument.
+  // Accepts _OneByteString as argument.
   // Returns index after last character written.
   int _setRange(int index, String oneByteString, int start, int end) {
     assert(oneByteString._isOneByte);
@@ -1320,6 +1372,7 @@ final class _OneByteString extends _StringBase {
   }
 }
 
+@pragma('vm:deeply-immutable')
 @pragma("vm:entry-point")
 final class _TwoByteString extends _StringBase {
   factory _TwoByteString._uninstantiable() {
@@ -1336,7 +1389,10 @@ final class _TwoByteString extends _StringBase {
 
   @pragma("vm:external-name", "TwoByteString_allocateFromTwoByteList")
   external static String _allocateFromTwoByteList(
-      List<int> list, int start, int end);
+    List<int> list,
+    int start,
+    int end,
+  );
 
   // This is internal helper method. Code point value must be a valid
   // UTF-16 value (0..0xFFFF), index must be valid.
@@ -1349,11 +1405,6 @@ final class _TwoByteString extends _StringBase {
     return _StringBase._isTwoByteWhitespace(codeUnit);
   }
 
-  @pragma("vm:recognized", "graph-intrinsic")
-  @pragma("vm:exact-result-type", "dart:core#_Smi")
-  @pragma("vm:external-name", "String_codeUnitAt")
-  external int codeUnitAt(int index);
-
   @pragma("vm:recognized", "asm-intrinsic")
   @pragma("vm:exact-result-type", bool)
   // Intrinsic is more efficient than an inlined body even for the small
@@ -1382,46 +1433,6 @@ final class _TwoByteString extends _StringBase {
   }
 }
 
-@pragma("vm:entry-point")
-final class _ExternalOneByteString extends _StringBase {
-  factory _ExternalOneByteString._uninstantiable() {
-    throw "Unreachable";
-  }
-
-  bool _isWhitespace(int codeUnit) {
-    return _StringBase._isOneByteWhitespace(codeUnit);
-  }
-
-  @pragma("vm:recognized", "graph-intrinsic")
-  @pragma("vm:exact-result-type", "dart:core#_Smi")
-  @pragma("vm:external-name", "String_codeUnitAt")
-  external int codeUnitAt(int index);
-
-  bool operator ==(Object other) {
-    return super == other;
-  }
-}
-
-@pragma("vm:entry-point")
-final class _ExternalTwoByteString extends _StringBase {
-  factory _ExternalTwoByteString._uninstantiable() {
-    throw "Unreachable";
-  }
-
-  bool _isWhitespace(int codeUnit) {
-    return _StringBase._isTwoByteWhitespace(codeUnit);
-  }
-
-  @pragma("vm:recognized", "graph-intrinsic")
-  @pragma("vm:exact-result-type", "dart:core#_Smi")
-  @pragma("vm:external-name", "String_codeUnitAt")
-  external int codeUnitAt(int index);
-
-  bool operator ==(Object other) {
-    return super == other;
-  }
-}
-
 final class _StringMatch implements Match {
   const _StringMatch(this.start, this.input, this.pattern);
 
@@ -1431,7 +1442,7 @@ final class _StringMatch implements Match {
 
   String group(int group) {
     if (group != 0) {
-      throw new RangeError.value(group);
+      throw RangeError.value(group);
     }
     return pattern;
   }
@@ -1457,12 +1468,12 @@ final class _StringAllMatchesIterable extends Iterable<Match> {
   _StringAllMatchesIterable(this._input, this._pattern, this._index);
 
   Iterator<Match> get iterator =>
-      new _StringAllMatchesIterator(_input, _pattern, _index);
+      _StringAllMatchesIterator(_input, _pattern, _index);
 
   Match get first {
     int index = _input.indexOf(_pattern, _index);
     if (index >= 0) {
-      return new _StringMatch(index, _input, _pattern);
+      return _StringMatch(index, _input, _pattern);
     }
     throw IterableElementError.noElement();
   }
@@ -1488,7 +1499,7 @@ final class _StringAllMatchesIterator implements Iterator<Match> {
       return false;
     }
     int end = index + _pattern.length;
-    _current = new _StringMatch(index, _input, _pattern);
+    _current = _StringMatch(index, _input, _pattern);
     // Empty match, don't start at same location again.
     if (end == _index) end++;
     _index = end;
