@@ -7,6 +7,12 @@ library dart.js_util;
 import "dart:_internal";
 import "dart:_js_helper";
 import "dart:_js_types";
+import "dart:js_interop"
+    show
+        JSAnyUtilityExtension,
+        FunctionToJSExportedDartFunction,
+        dartify,
+        JSAny;
 import "dart:_wasm";
 import "dart:async" show Completer, FutureOr;
 import "dart:collection";
@@ -20,22 +26,26 @@ dynamic jsify(Object? object) {
       return convertedObjects[o];
     }
 
+    // TODO(srujzs): We do these checks again in `jsifyRaw`. We should refactor
+    // this code so we don't have to, but we have to be careful about the
+    // `Iterable` check below.
     if (o == null ||
         o is num ||
         o is bool ||
         o is JSValue ||
         o is String ||
-        o is Int8List ||
-        o is Uint8List ||
-        o is Uint8ClampedList ||
-        o is Int16List ||
-        o is Uint16List ||
-        o is Int32List ||
-        o is Uint32List ||
-        o is Float32List ||
-        o is Float64List ||
-        o is ByteBuffer ||
-        o is ByteData) {
+        (o is TypedData &&
+            (o is Int8List ||
+                o is Uint8List ||
+                o is Uint8ClampedList ||
+                o is Int16List ||
+                o is Uint16List ||
+                o is Int32List ||
+                o is Uint32List ||
+                o is Float32List ||
+                o is Float64List ||
+                o is ByteData)) ||
+        o is ByteBuffer) {
       return JSValue(jsifyRaw(o));
     }
 
@@ -44,8 +54,11 @@ dynamic jsify(Object? object) {
       convertedObjects[o] = convertedMap;
       for (final key in o.keys) {
         final convertedKey = convert(key) as JSValue?;
-        setPropertyRaw(convertedMap.toExternRef, convertedKey?.toExternRef,
-            (convert(o[key]) as JSValue?)?.toExternRef);
+        setPropertyRaw(
+          convertedMap.toExternRef,
+          convertedKey.toExternRef,
+          (convert(o[key]) as JSValue?).toExternRef,
+        );
       }
       return convertedMap;
     } else if (o is Iterable<Object?>) {
@@ -86,8 +99,11 @@ T setProperty<T>(Object o, Object name, T? value) =>
         as T;
 
 @patch
-T callMethod<T>(Object o, Object method, List<Object?> args) => dartifyRaw(
-    callMethodVarArgsRaw(jsifyRaw(o), jsifyRaw(method), jsifyRaw(args))) as T;
+T callMethod<T>(Object o, Object method, List<Object?> args) =>
+    dartifyRaw(
+          callMethodVarArgsRaw(jsifyRaw(o), jsifyRaw(method), jsifyRaw(args)),
+        )
+        as T;
 
 @patch
 bool instanceof(Object? o, Object type) =>
@@ -151,23 +167,27 @@ typedef _PromiseFailureFunc = void Function(Object? error);
 Future<T> promiseToFuture<T>(Object jsPromise) {
   Completer<T> completer = Completer<T>();
 
-  final success = allowInterop<_PromiseSuccessFunc>((r) {
-    return completer.complete(r as FutureOr<T>?);
-  });
-  final error = allowInterop<_PromiseFailureFunc>((e) {
-    // Note that `completeError` expects a non-nullable error regardless of
-    // whether null-safety is enabled, so a `NullRejectionException` is always
-    // provided if the error is `null` or `undefined`.
-    // TODO(joshualitt): At this point `undefined` has been replaced with `null`
-    // so we cannot tell them apart. In the future we should reify `undefined`
-    // in Dart.
-    if (e == null) {
-      return completer.completeError(NullRejectionException(false));
-    }
-    return completer.completeError(e);
-  });
+  final success =
+      ((JSAny? jsValue) {
+        final r = dartifyRaw(jsValue.toExternRef);
+        return completer.complete(r as FutureOr<T>?);
+      }).toJS;
+  final error =
+      ((JSAny? jsError) {
+        // Note that `completeError` expects a non-nullable error regardless of
+        // whether null-safety is enabled, so a `NullRejectionException` is always
+        // provided if the error is `null` or `undefined`.
+        // TODO(joshualitt): At this point `undefined` has been replaced with `null`
+        // so we cannot tell them apart. In the future we should reify `undefined`
+        // in Dart.
+        final e = dartifyRaw(jsError.toExternRef);
+        if (e == null) {
+          return completer.completeError(NullRejectionException(false));
+        }
+        return completer.completeError(e);
+      }).toJS;
 
-  promiseThen(jsifyRaw(jsPromise), jsifyRaw(success), jsifyRaw(error));
+  promiseThen(jsifyRaw(jsPromise), success.toExternRef, error.toExternRef);
   return completer.future;
 }
 
@@ -179,8 +199,7 @@ Object? get objectPrototype => throw 'unimplemented';
 
 @patch
 List<Object?> objectKeys(Object? o) =>
-    dartifyRaw(JS<WasmExternRef?>('o => Object.keys(o)', jsifyRaw(o)))
-        as List<Object?>;
+    toDartList(JS<WasmExternRef?>('o => Object.keys(o)', jsifyRaw(o)));
 
 @patch
 Object? dartify(Object? object) {
@@ -189,7 +208,6 @@ Object? dartify(Object? object) {
     if (convertedObjects.containsKey(o)) {
       return convertedObjects[o];
     }
-
     // Because [List] needs to be shallowly converted across the interop
     // boundary, we have to double check for the case where a shallowly
     // converted [List] is passed back into [dartify].
@@ -200,37 +218,11 @@ Object? dartify(Object? object) {
       }
       return converted;
     }
-
-    if (o is! JSValue) {
-      return o;
-    }
-
+    if (o is! JSValue) return o;
     WasmExternRef? ref = o.toExternRef;
-    if (ref.isNull ||
-        isJSBoolean(ref) ||
-        isJSNumber(ref) ||
-        isJSString(ref) ||
-        isJSUndefined(ref) ||
-        isJSBoolean(ref) ||
-        isJSNumber(ref) ||
-        isJSString(ref) ||
-        isJSInt8Array(ref) ||
-        isJSUint8Array(ref) ||
-        isJSUint8ClampedArray(ref) ||
-        isJSInt16Array(ref) ||
-        isJSUint16Array(ref) ||
-        isJSInt32Array(ref) ||
-        isJSUint32Array(ref) ||
-        isJSFloat32Array(ref) ||
-        isJSFloat64Array(ref) ||
-        isJSArrayBuffer(ref) ||
-        isJSDataView(ref)) {
-      return dartifyRaw(ref);
-    }
-
+    final refType = externRefType(ref);
     // TODO(joshualitt) handle Date and Promise.
-
-    if (isJSSimpleObject(ref)) {
+    if (refType == ExternRefType.unknown && isJSSimpleObject(ref)) {
       final dartMap = <Object?, Object?>{};
       convertedObjects[o] = dartMap;
       // Keys will be a list of Dart [String]s.
@@ -238,22 +230,23 @@ Object? dartify(Object? object) {
       for (int i = 0; i < keys.length; i++) {
         final key = keys[i];
         if (key != null) {
-          dartMap[key] =
-              convert(JSValue.box(getPropertyRaw(ref, jsifyRaw(key))));
+          dartMap[key] = convert(
+            JSValue.box(getPropertyRaw(ref, jsifyRaw(key))),
+          );
         }
       }
       return dartMap;
-    } else if (isJSArray(ref)) {
+    }
+    if (refType == ExternRefType.array) {
       final dartList = <Object?>[];
       convertedObjects[o] = dartList;
       final length = getProperty<double>(o, 'length').toInt();
       for (int i = 0; i < length; i++) {
-        dartList.add(convert(JSValue.box(objectReadIndex(ref, i.toDouble()))));
+        dartList.add(convert(JSValue.box(objectReadIndex(ref, i))));
       }
       return dartList;
-    } else {
-      return dartifyRaw(ref);
     }
+    return dartifyRaw(ref, refType);
   }
 
   return convert(object);
