@@ -118,14 +118,33 @@ abstract mixin class Iterable<E> {
       Function id = _GeneratorIterable._id;
       if (id is! E Function(int)) {
         throw ArgumentError(
-            "Generator must be supplied or element type must allow integers",
-            "generator");
+          "Generator must be supplied or element type must allow integers",
+          "generator",
+        );
       }
       generator = id;
     }
     return _GeneratorIterable<E>(count, generator);
   }
 
+  /// Creates an [Iterable] from the [Iterator] factory.
+  ///
+  /// The returned iterable creates a new iterator each time [iterator] is read,
+  /// by calling the provided [iteratorFactory] function. The [iteratorFactory]
+  /// function must return a new instance of `Iterator<E>` on each call.
+  ///
+  /// This factory is useful when you need to create an iterable from a custom
+  /// iterator, or when you want to ensure a fresh iteration state on each use.
+  ///
+  /// Example:
+  /// ```dart
+  /// final numbers = Iterable.withIterator(() => [1, 2, 3].iterator);
+  /// print(numbers.toList()); // [1, 2, 3]
+  /// ```
+  @Since("3.8")
+  factory Iterable.withIterator(Iterator<E> Function() iteratorFactory) =
+      _WithIteratorIterable<E>;
+
   /// Creates an empty iterable.
   ///
   /// The empty iterable has no elements, and iterating it always stops
@@ -306,6 +325,7 @@ abstract mixin class Iterable<E> {
   /// This operation will check each element in order for being equal to
   /// [element], unless it has a more efficient way to find an element
   /// equal to [element].
+  /// Stops iterating on the first equal element.
   ///
   /// The equality used to determine whether [element] is equal to an element of
   /// the iterable defaults to the [Object.==] of the element.
@@ -368,6 +388,7 @@ abstract mixin class Iterable<E> {
   /// final result = numbers.reduce((value, element) => value + element);
   /// print(result); // 17.5
   /// ```
+  /// Consider using [fold] if the iterable can be empty.
   E reduce(E combine(E value, E element)) {
     Iterator<E> iterator = this.iterator;
     if (!iterator.moveNext()) {
@@ -411,6 +432,7 @@ abstract mixin class Iterable<E> {
   ///
   /// Checks every element in iteration order, and returns `false` if
   /// any of them make [test] return `false`, otherwise returns `true`.
+  /// Returns `true` if the iterable is empty.
   ///
   /// Example:
   /// ```dart
@@ -464,6 +486,7 @@ abstract mixin class Iterable<E> {
   ///
   /// Checks every element in iteration order, and returns `true` if
   /// any of them make [test] return `true`, otherwise returns false.
+  /// Returns `false` if the iterable is empty.
   ///
   /// Example:
   /// ```dart
@@ -508,7 +531,7 @@ abstract mixin class Iterable<E> {
   /// ```
   Set<E> toSet() => Set<E>.of(this);
 
-  /// The number of elements in [this].
+  /// The number of elements in this [Iterable].
   ///
   /// Counting all elements may involve iterating through all elements and can
   /// therefore be slow.
@@ -665,6 +688,7 @@ abstract mixin class Iterable<E> {
   /// Checks that this iterable has only one element, and returns that element.
   ///
   /// Throws a [StateError] if `this` is empty or has more than one element.
+  /// This operation will not iterate past the second element.
   E get single {
     Iterator<E> it = iterator;
     if (!it.moveNext()) throw IterableElementError.noElement();
@@ -689,6 +713,7 @@ abstract mixin class Iterable<E> {
   /// If no element satisfies [test], the result of invoking the [orElse]
   /// function is returned.
   /// If [orElse] is omitted, it defaults to throwing a [StateError].
+  /// Stops iterating on the first matching element.
   E firstWhere(bool test(E element), {E orElse()?}) {
     for (E element in this) {
       if (test(element)) return element;
@@ -799,8 +824,12 @@ abstract mixin class Iterable<E> {
       if (skipCount == 0) return iterator.current;
       skipCount--;
     }
-    throw IndexError.withLength(index, index - skipCount,
-        indexable: this, name: "index");
+    throw IndexError.withLength(
+      index,
+      index - skipCount,
+      indexable: this,
+      name: "index",
+    );
   }
 
   /// Returns a string representation of (some of) the elements of `this`.
@@ -825,8 +854,11 @@ abstract mixin class Iterable<E> {
   ///
   /// Handles circular references where converting one of the elements
   /// to a string ends up converting [iterable] to a string again.
-  static String iterableToShortString(Iterable iterable,
-      [String leftDelimiter = '(', String rightDelimiter = ')']) {
+  static String iterableToShortString(
+    Iterable iterable, [
+    String leftDelimiter = '(',
+    String rightDelimiter = ')',
+  ]) {
     if (isToStringVisiting(iterable)) {
       if (leftDelimiter == "(" && rightDelimiter == ")") {
         // Avoid creating a new string in the "common" case.
@@ -858,8 +890,11 @@ abstract mixin class Iterable<E> {
   ///
   /// Handles circular references where converting one of the elements
   /// to a string ends up converting [iterable] to a string again.
-  static String iterableToFullString(Iterable iterable,
-      [String leftDelimiter = '(', String rightDelimiter = ')']) {
+  static String iterableToFullString(
+    Iterable iterable, [
+    String leftDelimiter = '(',
+    String rightDelimiter = ')',
+  ]) {
     if (isToStringVisiting(iterable)) {
       return "$leftDelimiter...$rightDelimiter";
     }
@@ -898,6 +933,20 @@ class _GeneratorIterable<E> extends ListIterable<E> {
   static int _id(int n) => n;
 }
 
+/// Used internally by [Iterable.withIterator].
+///
+/// This class implements [Iterable] by delegating the iterator creation
+/// to the provided function.
+class _WithIteratorIterable<E> extends Iterable<E> {
+  final Iterator<E> Function() _iteratorFactory;
+
+  /// Creates an iterable that gets its elements from iterators created by
+  /// the provided factory function.
+  _WithIteratorIterable(this._iteratorFactory);
+
+  Iterator<E> get iterator => _iteratorFactory();
+}
+
 /// Convert elements of [iterable] to strings and store them in [parts].
 void _iterablePartsToStrings(Iterable<Object?> iterable, List<String> parts) {
   // This is the complicated part of [iterableToShortString].
