@@ -20,21 +20,26 @@ part of dart.core;
 ///
 /// It is generally not allowed to modify the map (add or remove keys) while
 /// an operation is being performed on the map, for example in functions called
-/// during a [forEach] or [putIfAbsent] call.
+/// during a [forEach] call.
 /// Modifying the map while iterating the keys or values
 /// may also break the iteration.
 ///
 /// It is generally not allowed to modify the equality of keys (and thus not
 /// their hashcode) while they are in the map. Some specialized subtypes may be
 /// more permissive, in which case they should document this behavior.
-abstract class Map<K, V> {
+///
+/// Key equality must be an equality relation. If the key stored in a map
+/// and the key used for lookup do not agree on whether the two are equal,
+/// so equality is not *symmetric*, then lookup behavior is unspecified.
+abstract interface class Map<K, V> {
   /// Creates an empty [LinkedHashMap].
   ///
-  /// This constructor is equivalent to the non-const map literal `<K,V>{}`.
+  /// This constructor is equivalent to the non-const map literal `<K, V>{}`.
   ///
   /// A `LinkedHashMap` requires the keys to implement compatible
   /// `operator==` and `hashCode`.
   /// It iterates in key insertion order.
+  // TODO: @Deprecated("Use literal <K, V>{} instead")
   external factory Map();
 
   /// Creates a [LinkedHashMap] with the same keys and values as [other].
@@ -51,6 +56,11 @@ abstract class Map<K, V> {
   /// A `LinkedHashMap` requires the keys to implement compatible
   /// `operator==` and `hashCode`.
   /// It iterates in key insertion order.
+  /// ```dart
+  /// final planets = <num, String>{1: 'Mercury', 2: 'Venus', 3: 'Earth', 4: 'Mars'};
+  /// final mapFrom = Map<int, String>.from(planets);
+  /// print(mapFrom); // {1: Mercury, 2: Venus, 3: Earth, 4: Mars}
+  /// ```
   factory Map.from(Map other) = LinkedHashMap<K, V>.from;
 
   /// Creates a [LinkedHashMap] with the same keys and values as [other].
@@ -58,6 +68,11 @@ abstract class Map<K, V> {
   /// A `LinkedHashMap` requires the keys to implement compatible
   /// `operator==` and `hashCode`, and it allows `null` as a key.
   /// It iterates in key insertion order.
+  /// ```dart
+  /// final planets = <int, String>{1: 'Mercury', 2: 'Venus', 3: 'Earth'};
+  /// final mapOf = Map<num, String>.of(planets);
+  /// print(mapOf); // {1: Mercury, 2: Venus, 3: Earth}
+  /// ```
   factory Map.of(Map<K, V> other) = LinkedHashMap<K, V>.of;
 
   /// Creates an unmodifiable hash-based map containing the entries of [other].
@@ -72,6 +87,11 @@ abstract class Map<K, V> {
   ///
   /// The resulting map behaves like the result of [Map.from],
   /// except that the map returned by this constructor is not modifiable.
+  /// ```dart
+  /// final planets = <int, String>{1: 'Mercury', 2: 'Venus', 3: 'Earth'};
+  /// final unmodifiableMap = Map.unmodifiable(planets);
+  /// unmodifiableMap[4] = 'Mars'; // Throws
+  /// ```
   external factory Map.unmodifiable(Map<dynamic, dynamic> other);
 
   /// Creates an identity map with the default implementation, [LinkedHashMap].
@@ -100,11 +120,11 @@ abstract class Map<K, V> {
   /// The keys of `map` are the `list` values converted to strings,
   /// and the values of the `map` are the squares of the `list` values:
   /// ```dart
-  /// List<int> list = [1, 2, 3];
-  /// var map = Map<String, int>.fromIterable(list,
+  /// final numbers = <int>[1, 2, 3];
+  /// final map = Map<String, int>.fromIterable(numbers,
   ///     key: (item) => item.toString(),
   ///     value: (item) => item * item);
-  /// // map is {"1": 1, "2": 4, "3": 9}
+  /// print(map); // {1: 1, 2: 4, 3: 9}
   /// ```
   /// If no values are specified for [key] and [value],
   /// the default is the identity function.
@@ -114,8 +134,9 @@ abstract class Map<K, V> {
   /// In the following example, the keys and corresponding values of `map`
   /// are the `list` values directly:
   /// ```dart
-  /// var map = Map<int, int>.fromIterable(list);
-  /// // map is {1: 1, 2: 2, 3: 3}
+  /// final numbers = <int>[1, 2, 3];
+  /// final map = Map.fromIterable(numbers);
+  /// print(map); // {1: 1, 2: 2, 3: 3}
   /// ```
   /// The keys computed by the source [iterable] do not need to be unique.
   /// The last occurrence of a key will overwrite
@@ -125,19 +146,21 @@ abstract class Map<K, V> {
   /// A `LinkedHashMap` requires the keys to implement compatible
   /// `operator==` and `hashCode`.
   /// It iterates in key insertion order.
-  factory Map.fromIterable(Iterable iterable,
-      {K key(dynamic element)?,
-      V value(dynamic element)?}) = LinkedHashMap<K, V>.fromIterable;
+  factory Map.fromIterable(
+    Iterable iterable, {
+    K key(dynamic element)?,
+    V value(dynamic element)?,
+  }) = LinkedHashMap<K, V>.fromIterable;
 
   /// Creates a map associating the given [keys] to the given [values].
   ///
   /// The map construction iterates over [keys] and [values] simultaneously,
   /// and adds an entry to the map for each pair of key and value.
   /// ```dart
-  /// List<String> letters = ['b', 'c'];
-  /// List<String> words = ['bad', 'cat'];
-  /// var map = Map.fromIterables(letters, words);
-  /// // map is {"b": "bad", "c": "cat"}
+  /// final rings = <bool>[false, false, true, true];
+  /// final planets = <String>{'Earth', 'Mars', 'Jupiter', 'Saturn'};
+  /// final map = Map<String, bool>.fromIterables(planets, rings);
+  /// print(map); // {Earth: false, Mars: false, Jupiter: true, Saturn: true}
   /// ```
   /// If [keys] contains the same object multiple times,
   /// the value of the last occurrence overwrites any previous value.
@@ -160,9 +183,14 @@ abstract class Map<K, V> {
   /// the store will throw unless the key is also an instance of [K] and
   /// the value is also an instance of [V].
   ///
-  /// If all accessed entries of [source] are have [K2] keys and [V2] values
-  /// and if all entries added to the returned map have [K] keys and [V]] values,
+  /// If all accessed entries of [source] have [K2] keys and [V2] values,
+  /// and if all entries added to the returned map have [K] keys and [V] values,
   /// then the returned map can be used as a `Map<K2, V2>`.
+  ///
+  /// Methods which accept `Object?` as argument,
+  /// like [containsKey], [remove] and [operator []],
+  /// will pass the argument directly to the this map's method
+  /// without any checks.
   static Map<K2, V2> castFrom<K, V, K2, V2>(Map<K, V> source) =>
       CastMap<K, V, K2, V2>(source);
 
@@ -178,6 +206,12 @@ abstract class Map<K, V> {
   /// ```dart
   /// <K, V>{for (var e in entries) e.key: e.value}
   /// ```
+  /// Example:
+  /// ```dart
+  /// final moonCount = <String, int>{'Mercury': 0, 'Venus': 0, 'Earth': 1,
+  ///   'Mars': 2, 'Jupiter': 79, 'Saturn': 82, 'Uranus': 27, 'Neptune': 14};
+  /// final map = Map.fromEntries(moonCount.entries);
+  /// ```
   factory Map.fromEntries(Iterable<MapEntry<K, V>> entries) =>
       <K, V>{}..addEntries(entries);
 
@@ -193,18 +227,37 @@ abstract class Map<K, V> {
   ///
   /// Entries added to the map must be valid for both a `Map<K, V>` and a
   /// `Map<RK, RV>`.
+  ///
+  /// Methods which accept `Object?` as argument,
+  /// like [containsKey], [remove] and [operator []],
+  /// will pass the argument directly to the this map's method
+  /// without any checks.
+  /// That means that you can do `mapWithStringKeys.cast<int,int>().remove("a")`
+  /// successfully, even if it looks like it shouldn't have any effect.
   Map<RK, RV> cast<RK, RV>();
 
   /// Whether this map contains the given [value].
   ///
   /// Returns true if any of the values in the map are equal to `value`
   /// according to the `==` operator.
+  /// ```dart
+  /// final moonCount = <String, int>{'Mercury': 0, 'Venus': 0, 'Earth': 1,
+  ///   'Mars': 2, 'Jupiter': 79, 'Saturn': 82, 'Uranus': 27, 'Neptune': 14};
+  /// final moons3 = moonCount.containsValue(3); // false
+  /// final moons82 = moonCount.containsValue(82); // true
+  /// ```
   bool containsValue(Object? value);
 
   /// Whether this map contains the given [key].
   ///
   /// Returns true if any of the keys in the map are equal to `key`
   /// according to the equality used by the map.
+  /// ```dart
+  /// final moonCount = <String, int>{'Mercury': 0, 'Venus': 0, 'Earth': 1,
+  ///   'Mars': 2, 'Jupiter': 79, 'Saturn': 82, 'Uranus': 27, 'Neptune': 14};
+  /// final containsUranus = moonCount.containsKey('Uranus'); // true
+  /// final containsPluto = moonCount.containsKey('Pluto'); // false
+  /// ```
   bool containsKey(Object? key);
 
   /// The value for the given [key], or `null` if [key] is not in the map.
@@ -222,7 +275,7 @@ abstract class Map<K, V> {
   /// Otherwise the key/value pair is added to the map.
   void operator []=(K key, V value);
 
-  /// The map entries of [this].
+  /// The map entries of this [Map].
   Iterable<MapEntry<K, V>> get entries;
 
   /// Returns a new map where all entries of this map are transformed by
@@ -236,28 +289,62 @@ abstract class Map<K, V> {
   ///
   /// The operation is equivalent to doing `this[entry.key] = entry.value`
   /// for each [MapEntry] of the iterable.
+  /// ```dart
+  /// final planets = <int, String>{1: 'Mercury', 2: 'Venus',
+  ///   3: 'Earth', 4: 'Mars'};
+  /// final gasGiants = <int, String>{5: 'Jupiter', 6: 'Saturn'};
+  /// final iceGiants = <int, String>{7: 'Uranus', 8: 'Neptune'};
+  /// planets.addEntries(gasGiants.entries);
+  /// planets.addEntries(iceGiants.entries);
+  /// print(planets);
+  /// // {1: Mercury, 2: Venus, 3: Earth, 4: Mars, 5: Jupiter, 6: Saturn,
+  /// //  7: Uranus, 8: Neptune}
+  /// ```
   void addEntries(Iterable<MapEntry<K, V>> newEntries);
 
   /// Updates the value for the provided [key].
   ///
   /// Returns the new value associated with the key.
   ///
-  /// If the key is present, invokes [update] with the current value and stores
-  /// the new value in the map.
+  /// If the key is present, invokes [update] with the current value
+  /// and stores the new value in the map.
   ///
   /// If the key is not present and [ifAbsent] is provided, calls [ifAbsent]
   /// and adds the key with the returned value to the map.
   ///
   /// If the key is not present, [ifAbsent] must be provided.
+  /// ```dart
+  /// final planetsFromSun = <int, String>{1: 'Mercury', 2: 'unknown',
+  ///   3: 'Earth'};
+  /// // Update value for known key value 2.
+  /// planetsFromSun.update(2, (value) => 'Venus');
+  /// print(planetsFromSun); // {1: Mercury, 2: Venus, 3: Earth}
+  ///
+  /// final largestPlanets = <int, String>{1: 'Jupiter', 2: 'Saturn',
+  ///   3: 'Neptune'};
+  /// // Key value 8 is missing from list, add it using [ifAbsent].
+  /// largestPlanets.update(8, (value) => 'New', ifAbsent: () => 'Mercury');
+  /// print(largestPlanets); // {1: Jupiter, 2: Saturn, 3: Neptune, 8: Mercury}
+  /// ```
   V update(K key, V update(V value), {V ifAbsent()?});
 
   /// Updates all values.
   ///
   /// Iterates over all entries in the map and updates them with the result
   /// of invoking [update].
+  /// ```dart
+  /// final terrestrial = <int, String>{1: 'Mercury', 2: 'Venus', 3: 'Earth'};
+  /// terrestrial.updateAll((key, value) => value.toUpperCase());
+  /// print(terrestrial); // {1: MERCURY, 2: VENUS, 3: EARTH}
+  /// ```
   void updateAll(V update(K key, V value));
 
   /// Removes all entries of this map that satisfy the given [test].
+  /// ```dart
+  /// final terrestrial = <int, String>{1: 'Mercury', 2: 'Venus', 3: 'Earth'};
+  /// terrestrial.removeWhere((key, value) => value.startsWith('E'));
+  /// print(terrestrial); // {1: Mercury, 2: Venus}
+  /// ```
   void removeWhere(bool test(K key, V value));
 
   /// Look up the value of [key], or add a new entry if it isn't there.
@@ -265,16 +352,30 @@ abstract class Map<K, V> {
   /// Returns the value associated to [key], if there is one.
   /// Otherwise calls [ifAbsent] to get a new value, associates [key] to
   /// that value, and then returns the new value.
+  ///
+  /// That is, if the key is currently in the map,
+  /// `map.putIfAbsent(key, ifAbsent)` is equivalent to `map[key]`.
+  /// If the key is not currently in the map,
+  /// it's instead equivalent to `map[key] = ifAbsent()`
+  /// (but without any guarantee that the `[]` and `[]=` operators are
+  /// actually called to achieve that effect).
+  ///
   /// ```dart
-  /// Map<String, int> scores = {'Bob': 36};
-  /// for (var key in ['Bob', 'Rohan', 'Sophena']) {
-  ///   scores.putIfAbsent(key, () => key.length);
+  /// final diameters = <num, String>{1.0: 'Earth'};
+  /// final otherDiameters = <double, String>{0.383: 'Mercury', 0.949: 'Venus'};
+  ///
+  /// for (final item in otherDiameters.entries) {
+  ///   diameters.putIfAbsent(item.key, () => item.value);
   /// }
-  /// scores['Bob'];      // 36
-  /// scores['Rohan'];    //  5
-  /// scores['Sophena'];  //  7
+  /// print(diameters); // {1.0: Earth, 0.383: Mercury, 0.949: Venus}
+  ///
+  /// // If the key already exists, the current value is returned.
+  /// final result = diameters.putIfAbsent(0.383, () => 'Random');
+  /// print(result); // Mercury
+  /// print(diameters); // {1.0: Earth, 0.383: Mercury, 0.949: Venus}
   /// ```
-  /// Calling [ifAbsent] must not add or remove keys from the map.
+  /// The [ifAbsent] function is allowed to modify the map,
+  /// and if so, it behaves the same as the equivalent `map[key] = ifAbsent()`.
   V putIfAbsent(K key, V ifAbsent());
 
   /// Adds all key/value pairs of [other] to this map.
@@ -284,6 +385,11 @@ abstract class Map<K, V> {
   /// The operation is equivalent to doing `this[key] = value` for each key
   /// and associated value in other. It iterates over [other], which must
   /// therefore not change during the iteration.
+  /// ```dart
+  /// final planets = <int, String>{1: 'Mercury', 2: 'Earth'};
+  /// planets.addAll({5: 'Jupiter', 6: 'Saturn'});
+  /// print(planets); // {1: Mercury, 2: Earth, 5: Jupiter, 6: Saturn}
+  /// ```
   void addAll(Map<K, V> other);
 
   /// Removes [key] and its associated value, if present, from the map.
@@ -293,19 +399,40 @@ abstract class Map<K, V> {
   ///
   /// Note that some maps allow `null` as a value,
   /// so a returned `null` value doesn't always mean that the key was absent.
+  /// ```dart
+  /// final terrestrial = <int, String>{1: 'Mercury', 2: 'Venus', 3: 'Earth'};
+  /// final removedValue = terrestrial.remove(2); // Venus
+  /// print(terrestrial); // {1: Mercury, 3: Earth}
+  /// ```
   V? remove(Object? key);
 
   /// Removes all entries from the map.
   ///
   /// After this, the map is empty.
+  /// ```dart
+  /// final planets = <int, String>{1: 'Mercury', 2: 'Venus', 3: 'Earth'};
+  /// planets.clear(); // {}
+  /// ```
   void clear();
 
   /// Applies [action] to each key/value pair of the map.
   ///
   /// Calling `action` must not add or remove keys from the map.
+  /// ```dart
+  /// final planetsByMass = <num, String>{0.81: 'Venus', 1: 'Earth',
+  ///   0.11: 'Mars', 17.15: 'Neptune'};
+  ///
+  /// planetsByMass.forEach((key, value) {
+  ///   print('$key: $value');
+  ///   // 0.81: Venus
+  ///   // 1: Earth
+  ///   // 0.11: Mars
+  ///   // 17.15: Neptune
+  /// });
+  /// ```
   void forEach(void action(K key, V value));
 
-  /// The keys of [this].
+  /// The keys of this [Map].
   ///
   /// The returned iterable has efficient `length` and `contains` operations,
   /// based on [length] and [containsKey] of the map.
@@ -316,7 +443,7 @@ abstract class Map<K, V> {
   /// Modifying the map while iterating the keys may break the iteration.
   Iterable<K> get keys;
 
-  /// The values of [this].
+  /// The values of this [Map].
   ///
   /// The values are iterated in the order of their corresponding keys.
   /// This means that iterating [keys] and [values] in parallel will
@@ -340,11 +467,40 @@ abstract class Map<K, V> {
 }
 
 /// A key/value pair representing an entry in a [Map].
-class MapEntry<K, V> {
+///
+/// The [Map] interface contains various methods that can
+/// inspect or modify the map based on entry objects.
+/// ```dart
+/// final map = {'1': 'A', '2': 'B'};
+/// map.addEntries([
+///  MapEntry('3', 'C'),
+///  MapEntry('4', 'D'),
+/// ]);
+/// print(map); // {1: A, 2: B, 3: C, 4: D}
+/// ```
+///
+/// Do not extend or implement the `MapEntry` class.
+/// If the Dart language introduces value types,
+/// the `MapEntry` class will be changed to such a type,
+/// and will likely no longer be able to be implemented or extended
+/// by classes.
+final class MapEntry<K, V> {
   /// The key of the entry.
+  ///
+  /// ```dart
+  /// final map = {'theKey': 'theValue'}; // Map<String, String>
+  /// var entry = map.entries.first; // MapEntry<String, String>
+  /// print(entry.key); // 'theKey'
+  /// ```
   final K key;
 
-  /// The value associated to [key] in the map.
+  /// The value associated to [key] in a map.
+  ///
+  /// ```dart
+  /// final map = {'theKey': 'theValue'}; // Map<String, String>
+  /// var entry = map.entries.first; // MapEntry<String, String>
+  /// print(entry.value); // 'theValue'
+  /// ```
   final V value;
 
   /// Creates an entry with [key] and [value].
@@ -352,5 +508,8 @@ class MapEntry<K, V> {
 
   const MapEntry._(this.key, this.value);
 
+  /// String representation intended for debugging only.
+  ///
+  /// Not guaranteed to be stable over time.
   String toString() => "MapEntry($key: $value)";
 }
