@@ -123,10 +123,16 @@ abstract final class String implements Comparable<String>, Pattern {
   /// ```
   /// If [start] and [end] are provided, only the values of [charCodes]
   /// at positions from `start` to, but not including, `end`, are used.
-  /// The `start` and `end` values must satisfy
-  /// `0 <= start <= end <= charCodes.length`.
-  external factory String.fromCharCodes(Iterable<int> charCodes,
-      [int start = 0, int? end]);
+  /// The `start` and `end` values must satisfy `0 <= start <= end`.
+  /// If [start] is omitted, it defaults to zero, the start of [charCodes],
+  /// and if [end] is omitted, all char-codes after [start] are included.
+  /// If [charCodes] does not have [end], or even [start], elements,
+  /// the specified char-codes may be shorter than `end - start`, or even empty.
+  external factory String.fromCharCodes(
+    Iterable<int> charCodes, [
+    int start = 0,
+    int? end,
+  ]);
 
   /// Allocates a new string containing the specified [charCode].
   ///
@@ -138,20 +144,27 @@ abstract final class String implements Comparable<String>, Pattern {
   /// Creating a [String] with one half of a surrogate pair is allowed.
   external factory String.fromCharCode(int charCode);
 
-  /// The string value of the environment declaration [name].
+  /// Value for [name] in the compilation configuration environment declaration.
   ///
-  /// Environment declarations are provided by the surrounding system compiling
-  /// or running the Dart program. Declarations map a string key to a string
-  /// value.
+  /// The compilation configuration environment is provided by the
+  /// surrounding tools which are compiling or running the Dart program.
+  /// The environment is a mapping from a set of string keys to their associated
+  /// string value.
+  /// The string value, or lack of a value, associated with a [name]
+  /// must be consistent across all calls to `String.fromEnvironment`,
+  /// [int.fromEnvironment], [bool.fromEnvironment] and [bool.hasEnvironment]
+  /// in a single program.
   ///
-  /// If [name] is not declared in the environment, the result is instead
-  /// [defaultValue].
+  /// The result of invoking this constructor is the string associated with
+  /// the key [name]. If no value is associated with [name],
+  /// the result is instead the [defaultValue] string, which defaults to
+  /// the empty string.
   ///
-  /// Example of getting a value:
+  /// Example of looking up a value:
   /// ```dart
   /// const String.fromEnvironment("defaultFloo", defaultValue: "no floo")
   /// ```
-  /// In order to check whether a declaration is there at all, use
+  /// In order to check whether a value is there at all, use
   /// [bool.hasEnvironment]. Example:
   /// ```dart
   /// const maybeDeclared = bool.hasEnvironment("maybeDeclared")
@@ -168,14 +181,14 @@ abstract final class String implements Comparable<String>, Pattern {
   /// It may work as a non-constant invocation on some platforms which
   /// have access to compiler options at run-time, but most ahead-of-time
   /// compiled platforms will not have this information.
-  // The .fromEnvironment() constructors are special in that we do not want
-  // users to call them using "new". We prohibit that by giving them bodies
-  // that throw, even though const constructors are not allowed to have bodies.
-  // Disable those static errors.
-  //ignore: const_constructor_with_body
-  //ignore: const_factory
-  external const factory String.fromEnvironment(String name,
-      {String defaultValue = ""});
+  ///
+  /// The compilation configuration environment is not the same as the
+  /// environment variables of a POSIX process. Those can be accessed on
+  /// native platforms using `Platform.environment` from the `dart:io` library.
+  external const factory String.fromEnvironment(
+    String name, {
+    String defaultValue = "",
+  });
 
   /// The character (as a single-code-unit [String]) at the given [index].
   ///
@@ -514,8 +527,11 @@ abstract final class String implements Comparable<String>, Pattern {
   /// is replaced by the result of calling [replace] with the match object.
   ///
   /// The [startIndex] must be non-negative and no greater than [length].
-  String replaceFirstMapped(Pattern from, String replace(Match match),
-      [int startIndex = 0]);
+  String replaceFirstMapped(
+    Pattern from,
+    String replace(Match match), [
+    int startIndex = 0,
+  ]);
 
   /// Replaces all substrings that match [from] with [replace].
   ///
@@ -685,8 +701,11 @@ abstract final class String implements Comparable<String>, Pattern {
   ///     onNonMatch: (n) => '*');
   /// print(result); // *shoots*
   /// ```
-  String splitMapJoin(Pattern pattern,
-      {String Function(Match)? onMatch, String Function(String)? onNonMatch});
+  String splitMapJoin(
+    Pattern pattern, {
+    String Function(Match)? onMatch,
+    String Function(String)? onNonMatch,
+  });
 
   /// An unmodifiable list of the UTF-16 code units of this string.
   List<int> get codeUnits;
@@ -762,7 +781,7 @@ abstract final class String implements Comparable<String>, Pattern {
 /// https://dart.dev/guides/language/language-tour#runes-and-grapheme-clusters)
 /// in
 /// [A tour of the Dart language](https://dart.dev/guides/language/language-tour).
-class Runes extends Iterable<int> {
+final class Runes extends Iterable<int> {
   /// The string that this is the runes of.
   final String string;
 
@@ -799,7 +818,7 @@ int _combineSurrogatePair(int start, int end) {
 }
 
 /// [Iterator] for reading runes (integer Unicode code points) of a Dart string.
-class RuneIterator implements Iterator<int> {
+final class RuneIterator implements Iterator<int> {
   /// String being iterated.
   final String string;
 
@@ -817,9 +836,9 @@ class RuneIterator implements Iterator<int> {
 
   /// Create an iterator positioned at the beginning of the string.
   RuneIterator(String string)
-      : this.string = string,
-        _position = 0,
-        _nextPosition = 0;
+    : this.string = string,
+      _position = 0,
+      _nextPosition = 0;
 
   /// Create an iterator positioned before the [index]th code unit of the string.
   ///
@@ -830,9 +849,9 @@ class RuneIterator implements Iterator<int> {
   ///
   /// The [index] position must not be in the middle of a surrogate pair.
   RuneIterator.at(String string, int index)
-      : string = string,
-        _position = index,
-        _nextPosition = index {
+    : string = string,
+      _position = index,
+      _nextPosition = index {
     RangeError.checkValueInInterval(index, 0, string.length);
     _checkSplitSurrogate(index);
   }
@@ -861,8 +880,12 @@ class RuneIterator implements Iterator<int> {
   /// Setting the position to the end of the string means that there is no
   /// current rune.
   void set rawIndex(int rawIndex) {
-    IndexError.check(rawIndex, string.length,
-        indexable: string, name: "rawIndex");
+    IndexError.check(
+      rawIndex,
+      string.length,
+      indexable: string,
+      name: "rawIndex",
+    );
     reset(rawIndex);
     moveNext();
   }
